from abc import ABC, abstractmethod
import numpy as np
import numpy.matlib as matlib
import matplotlib.pyplot as plt
import math
from scipy.fftpack import dst, idst
import scipy.sparse as sparse
import operator
from functools import reduce
import warnings
from cuqi._messages import _disable_warning_msg

class Geometry(ABC):
    """A class that represents the geometry of the range, domain, observation, or other sets.

    It specifies a mapping from the parameter space to the function space (:meth:`par2fun`) and the inverse map if possible (:meth:`fun2par`). The parameters can be for example, the center and width of a hat function, and the function is the resulting hat function evaluated at grid points of a given grid. The geometry keeps track of the dimension and shape of the parameter space (:meth:`par_dim` and :meth:`par_shape`) and the dimension and shape of the function space (:meth:`fun_dim` and :meth:`fun_shape`).
    """
    @property
    @abstractmethod
    def par_shape(self):
        """The shape of the parameter space."""
        pass

    @property
    def par_dim(self):
        """The dimension of the geometry (parameter space)."""
        if self.par_shape is None: return None
        return reduce(operator.mul, self.par_shape) # math.prod(self.par_shape) for Python 3.8+

    @property
    def fun_shape(self):
        """The shape of the geometry (function space). """
        if not hasattr(self,'_fun_shape') or self._fun_shape is None:
            # Attempt to infer dimension
            funvals = self.par2fun(np.ones(self.par_dim))
            if hasattr(funvals, 'shape'):
                self._fun_shape = funvals.shape
            else:
                self._fun_shape = None
        return self._fun_shape
    
    @property
    def fun_dim(self):
        """The dimension of the geometry (function space). """
        if self.fun_shape is None: return None
        return reduce(operator.mul, self.fun_shape) # math.prod(self.fun_shape) for Python 3.8+

    @property
    def funvec_shape(self):
        """The shape of the geometry (shape of the vector representation of the
        function value)."""
        if not hasattr(self,'_funvec_shape') or self._funvec_shape is None:
            # Attempt to infer dimension
            funvecvals = self.fun2vec(self.par2fun(np.ones(self.par_dim)))
            if hasattr(funvecvals, 'shape'):
                self._funvec_shape = funvecvals.shape
                if len(self._funvec_shape) != 1:
                    raise ValueError("funvec_shape must be a 1D tuple.")
            else:
                warnings.warn("Could not infer function space shape.")
                self._funvec_shape = None
        return self._funvec_shape        

    @property
    def funvec_dim(self):
        """The dimension of the geometry (dimension of the vector representation
        of the function value). """
        if self.funvec_shape is None: return None
        return reduce(operator.mul, self.funvec_shape) 
    
    @property
    def variables(self):
        #No variable names set, generate variable names from dim
        if not hasattr(self,"_variables"):
                self.variables = self.par_dim
        return self._variables

    @variables.setter
    def variables(self, value):
        """ Set variable names with a list of strings. Generic variable names can also be generated by passing an integer. """
        variables_value_err_msg = "variables should be int, or list of strings"
        if isinstance(value,(int,np.integer)):
            # Find name of variable if set, else set it to a default
            if hasattr(self, "_variable_name") and self._variable_name is not None:
                name = self._variable_name
            else:
                name = "v"
            # If more than one variable we index, else we dont
            value = [name+str(var) for var in range(value)] if value != 1 else [name]
        elif isinstance(value,list): 
            for var in value: 
                if not isinstance(var,str):
                    raise ValueError(variables_value_err_msg)
        else:
            raise ValueError(variables_value_err_msg) 
        self._variables = value
        self._ids = range(self.par_dim)
    
    @property
    def fun_is_array(self):
        """Flag to indicate whether the function value is an array. This can be
        useful to query when converting samples to function values."""
        if isinstance(self.fun_shape, tuple) and \
           all([isinstance(i, int) for i in self.fun_shape]):
            return True
        else:
            return False

    
    def plot(self, values, is_par=True, plot_par=False, **kwargs):
        """
        Plots a function over the set defined by the geometry object.
            
        Parameters
        ----------
        values : ndarray
            1D array that contains the values of the function degrees of freedom.

        is_par : Boolean, default True
            Flag to indicate whether the values are parameters or function values.
            True:  values are passed through the :meth:`par2fun` method.
            False: values are plotted directly.
        
        plot_par : Boolean, default False
            If true this method plots the parameters as a :class:`Discrete` geometry.
        """
        #Error check
        if plot_par and not is_par:
            raise Exception("Plot par is true, but is_par is false (parameters were not given)")

        if plot_par:
            geom = Discrete(self.par_dim) #par_dim is size of the parameter space.
            return geom.plot(values, **kwargs)

        if is_par:
            values = self.par2fun(values)

        return self._plot(values, **kwargs)

    def plot_envelope(self, lo_values, hi_values, is_par=True, plot_par=False, **kwargs):
        """
        Plots an envelope from lower and upper bounds over the set defined by the geometry object.
            
        Parameters
        ----------
        lo_values : ndarray
            1D array that contains a lower bound of the function degrees of freedom.

        hi_values : ndarray
            1D array that contains an upper bound of the function degrees of freedom.

        is_par : Boolean, default True
            Flag to indicate whether the values are parameters or function values.
            True:  values are passed through the :meth:`par2fun` method.
            False: values are plotted directly.
        
        plot_par : Boolean, default False
            If true this method plots the parameters as a :class:`Discrete` geometry.
        """
        #Error check
        if plot_par and not is_par:
            raise ValueError("Plot par is true, but is_par is false (parameters were not given)")
        
        if plot_par:
            geom = Discrete(self.par_dim) #par_dim is size of the parameter space.
            return geom.plot_envelope(lo_values, hi_values, **kwargs)

        if is_par:
            lo_values = self.par2fun(lo_values)
            hi_values = self.par2fun(hi_values)

        return self._plot_envelope(lo_values,hi_values, **kwargs)

    def par2fun(self, par):
        """The parameter to function map used to map parameters to function values in e.g. plotting."""
        return par

    def fun2par(self, funvals):
        """The function to parameter map used to map function values back to parameters, if available."""
        raise NotImplementedError("fun2par is not implemented. Must be implemented specifically for each geometry.")

    def fun2vec(self, funvals):
        """Maps function values to a vector representation of the function values, if available."""
        if self.fun_is_array and len(self.fun_shape) == 1:
            return funvals
        else:
            raise NotImplementedError("fun2vec is not implemented. Must be implemented specifically for each geometry.")
    
    def vec2fun(self, funvec):
        """Maps function vector representation, if available, to function values."""
        if self.fun_is_array and len(self.fun_shape) == 1:
            return funvec
        raise NotImplementedError("vec2fun is not implemented. Must be implemented specifically for each geometry.")
    
    @abstractmethod
    def _plot(self):
        pass

    def _plot_envelope(self, *args, **kwargs):
        raise NotImplementedError("Plot envelope not implemented for {}. Use flag plot_par to plot envelope of parameters instead.".format(type(self)))
            
    def _plot_config(self,values):
        """
        A method that implements any default configuration for the plots. This method is to be called inside any 'plot_' method.
        """
        pass

    def _create_subplot_list(self,Ns,subplots=True):
        Nx = math.ceil(np.sqrt(Ns))
        Ny = Nx
        subplot_ids = []
        fig = plt.gcf()
        if subplots: fig.set_size_inches(fig.bbox_inches.corners()[3][0]*Nx, fig.bbox_inches.corners()[3][1]*Ny)

        for i in range(Ny):
            for j in range(Nx):
                subplot_id = i*Nx+j+1
                if subplot_id > Ns: 
                    continue 
                subplot_ids.append((Ny,Nx,subplot_id))
        return subplot_ids

    def __eq__(self, obj):
        if not isinstance(obj, self.__class__): return False
        return self._all_values_equal(obj)

    def __repr__(self) -> str:
        if self.par_shape == self.fun_shape:
            return "{}[{}]".format(self.__class__.__name__,
                                   self.par_shape if len(self.par_shape) != 1 else self.par_shape[0])
        return "{}[{}: {}]".format(
            self.__class__.__name__,
            self.par_shape if len(self.par_shape) != 1 else self.par_shape[0],
            (
                self.fun_shape
                if (self.fun_shape is None or len(self.fun_shape) != 1)
                else self.fun_shape[0]
            ),
        )

    def _all_values_equal(self, obj):
        """Returns true of all values of the object and self are equal"""
        for key, value in vars(self).items():
            # If _variables exist ensure it exists in both objects (by calling generator)
            if key == "_variables": obj.variables
            if key == "_variable_name" and not hasattr(obj, "_variable_name"): obj._variable_name = None

            # Store value to compare
            obj_value = vars(obj)[key]

            # If list/tuple we compare each element
            if isinstance(value, (tuple,list)) and isinstance(obj_value, (tuple,list)):
                for i in range(max(len(value),len(obj_value))):
                    if not np.array_equiv(value[i],obj_value[i]):
                        return False
            # Else we check single element
            elif not np.array_equiv(value, obj_value):
                return False 
        return True


class _WrappedGeometry(Geometry):
    """A class that wraps a given geometry (emulates dynamic
    inheritance from the given geometry).
    
    Parameters
    -----------
    geometry : cuqi.geometry.Geometry
    """

    def __init__(self,geometry):
        self.geometry = geometry

    @property
    def par_shape(self):
        return self.geometry.par_shape

    @property
    def grid(self):
        return self.geometry.grid

    @property
    def axis_labels(self):
        return self.geometry.axis_labels

    @property
    def variables(self):
        return self.geometry.variables

    @property
    def mesh(self):
        return self.geometry.mesh

    def par2fun(self,p):
        return self.geometry.par2fun(p)

    def fun2par(self,f):
        return self.geometry.fun2par(f)

    def _plot(self, values, *args, **kwargs):
        """Calls the underlying geometry plotting method."""
        return self.geometry._plot(values, *args, **kwargs)

    def _plot_envelope(self, lo_values, hi_values, *args, **kwargs):
        """Calls the underlying geometry plotting of envelope method."""
        return self.geometry._plot_envelope(lo_values, hi_values, *args, **kwargs)

    def _process_values(self,values):
        return self.geometry._process_values(values) 

    def __repr__(self) -> str:
        return "{} wraps {}".format(self.__class__.__name__,self.geometry.__repr__())


class Continuous(Geometry, ABC):

    def __init__(self,grid=None, axis_labels=None):
        self.axis_labels = axis_labels
        self.grid = grid

    def _create_dimension(self, dim_grid):
        dim_grid_value_err_msg = "dim_grid should be int, tuple with one int element, list of numbers, 1D numpy.ndarray, or None"
        if dim_grid is None:
            return None

        if isinstance(dim_grid,tuple) and len(dim_grid)==1:
            dim_grid = dim_grid[0]

        if isinstance(dim_grid,(int,np.integer)):
            dim_grid = np.arange(dim_grid)
        elif isinstance(dim_grid,(list,np.ndarray)):
            dim_grid = np.array(dim_grid)
            if len(dim_grid.shape)!=1:
                raise ValueError(dim_grid_value_err_msg)
        else:
            raise ValueError(dim_grid_value_err_msg)
        return dim_grid.astype(float)
        
    @property
    def grid(self):
        return self._grid
    
    def fun2par(self,funvals):
        return funvals
    
    def _reshape_par2fun_input(self, pars):
        """Make sure that the parameter last dimension reflects the number of
        parameter vectors: 1 for a single parameter vector and n for n parameter
        vectors."""
        # Ensure pars shape is correct (either squeezed single vector or 
        # multiple vectors with the correct shape).
        if pars.shape != self.par_shape and\
                pars.shape[:-1] != self.par_shape:
            raise ValueError(
                f"pars must have shape {self.par_shape} or {self.par_shape}"
                + "+(n,) where n is the number of parameter vectors")

        # Add dim to pars if pars is a single parameter vector.
        return pars.reshape(self.par_shape + (-1,))

    def _reshape_fun2par_input(self, funvals):
        """Make sure that the function value last dimension reflects the number
        of functions: 1 for a single function and n for n functions."""
        # Ensure funvals shape is correct (either squeezed single function or 
        # multiple functions with the correct shape).
        if funvals.shape != self.fun_shape and\
                funvals.shape[:-1] != self.fun_shape:
            raise ValueError(
                f"funvals must have shape {self.fun_shape} or {self.fun_shape}"
                + "+(n,) where n is the number of functions")

        # Add dim to funvals if funvals is a single function.
        return funvals.reshape(self.fun_shape + (-1,))


class Continuous1D(Continuous):
    """A class that represents a continuous 1D geometry.

    Parameters
    -----------
    grid : int, tuple, list or numpy.ndarray
        1D array of node coordinates in a 1D grid (list or numpy.ndarray), or number of nodes (int or tuple with one int element) in the grid. In the latter case, a default grid with unit spacing and coordinates 0,1,2,... will be created.

    Attributes
    -----------
    grid : numpy.ndarray
        1D array of node coordinates in a 1D grid
    """

    def __init__(self, grid=None, axis_labels=None, **kwargs):
        super().__init__(grid, axis_labels, **kwargs)

    @property
    def fun_shape(self):
        if self.grid is None: return None
        return self.grid.shape

    @property
    def par_shape(self):
        """The shape of the parameter space"""
        if self.grid is None: return None
        return (len(self.grid), )

    @Continuous.grid.setter
    def grid(self, value):
        self._grid = self._create_dimension(value)

    def _plot(self, values, *args, **kwargs):
        if self.par_dim==1 and "marker" not in kwargs.keys():
            kwargs["marker"] = "o"
        p = plt.plot(self.grid, values, *args, **kwargs)
        self._plot_config()
        return p

    def _plot_envelope(self, lo_values, up_values, **kwargs):
        default = {'color':'dodgerblue', 'alpha':0.25}
        for key in default:
            if (key not in kwargs.keys()):
                kwargs[key]  = default[key]
        return plt.fill_between(self.grid,up_values, lo_values, **kwargs)

    def _plot_config(self):
        if self.axis_labels is not None:
            plt.xlabel(self.axis_labels[0])


class Continuous2D(Continuous):

    def __init__(self,grid=None,axis_labels=None):
        super().__init__(grid, axis_labels)
            
    @property
    def fun_shape(self): 
        if self.grid is None: return None
        return (len(self.grid[0]), len(self.grid[1])) 

    @property
    def par_shape(self):
        """The shape of the parameter space"""
        if self.grid is None: return None
        return (len(self.grid[0])*len(self.grid[1]), ) 

    @Continuous.grid.setter
    def grid(self, value):
        if value is None: self._grid = None
        else:
            if len(value)!=2:
                raise NotImplementedError("grid must be a 2D tuple of int values or arrays (list, tuple or numpy.ndarray) or combination of both")
            self._grid = (self._create_dimension(value[0]), self._create_dimension(value[1]))

    def _plot(self, values, plot_type='pcolor', **kwargs):
        """
        Overrides :meth:`cuqi.geometry.Geometry.plot`. See :meth:`cuqi.geometry.Geometry.plot` for description  and definition of the parameter `values`.
        
        Parameters
        -----------
        plot_type : str
            type of the plot. If plot_type = 'pcolor', :meth:`matplotlib.pyplot.pcolor` is called, if plot_type = 'contour', :meth:`matplotlib.pyplot.contour` is called, and if `plot_type` = 'contourf', :meth:`matplotlib.pyplot.contourf` is called, 

        kwargs : keyword arguments
            keyword arguments which the methods :meth:`matplotlib.pyplot.pcolor`, :meth:`matplotlib.pyplot.contour`, or :meth:`matplotlib.pyplot.contourf`  normally take, depending on the value of the parameter `plot_type`.
        """
        if plot_type == 'pcolor': 
            plot_method = plt.pcolor
        elif plot_type == 'contour':
            plot_method = plt.contour
        elif plot_type == 'contourf':
            plot_method = plt.contourf
        else:
            raise ValueError(f"unknown value: {plot_type} of the parameter 'plot_type'")
        
        values = self._process_values(values)
        subplot_ids = self._create_subplot_list(values.shape[-1])
        ims = []
        for rows,cols,subplot_id in subplot_ids:
            plt.subplot(rows,cols,subplot_id); 
            ims.append(plot_method(self.grid[0], self.grid[1], values[..., subplot_id-1].reshape(self.fun_shape[::-1]),
                                   **kwargs))
        self._plot_config()
        return ims

    def plot_pcolor(self, values, **kwargs):
        return self.plot(values, plot_type='pcolor', **kwargs)

    def plot_contour(self, values, **kwargs):
        return self.plot(values, plot_type='contour', **kwargs)

    def plot_contourf(self, values, **kwargs):
       return self.plot(values, plot_type='contourf', **kwargs)
    
    def _process_values(self,values):
        if len(values.shape) == 3 or\
             (len(values.shape) == 2 and values.shape[0]== self.par_dim):  
            pass
        else:
            values = values[..., np.newaxis]
        return values

    def _plot_config(self):
        for i, axis in enumerate(plt.gcf().axes):
            if self.axis_labels is not None:
                axis.set_xlabel(self.axis_labels[0])
                axis.set_ylabel(self.axis_labels[1])
            axis.set_aspect('equal')

    def par2fun(self, pars):
        """Converts a parameter vector or multiple parameter vectors into a 2D
         function value or multiple 2D function values."""
        # Reshape to 2D domain (also for multiple parameter vectors, e.g.
        # Samples.samples).
        funvals = pars.reshape(self.fun_shape+(-1,))
        # Squeeze to return single evaluated function if only one parameter
        # vector was given
        return funvals.squeeze()

    def fun2par(self, funvals):
        """Converts a 2D function value or multiple 2D function values into a
        parameter vector or multiple parameter vectors."""
        # Reshape to parameter vector (also for multiple function values,
        # e.g. Samples.funvals.samples).
        pars = funvals.reshape((self.par_dim,)+(-1,))
        # Squeeze to return single parameter vector if only one 2D function
        # value was given
        return pars.squeeze()


class Image2D(Geometry):
    """ A class that represents a 2D image.

    The par2fun method converts the parameter vector into an image (matrix).
    The fun2par method converts the image (matrix) into a parameter vector.

    Plotting is handled via matplotlib.pyplot.imshow.
    Colormap is defaulted to grayscale.

    A How-To guide on the use of Image2D as the domain/range geometry of a CUQI 
    :class:`Model` is available `here <https://cuqi-dtu.github.io/CUQIpy/user/_auto_howtos/Image2D.html>`_.

    Parameters
    -----------
    im_shape : tuple
        shape of the image (rows, columns)

    order : str
        If order = 'C', the image is represented in row-major order.
        if order = 'F', the image is represented column-major order.

    visual_only : bool, Default: False
        If visual_only = True, par2fun and fun2par will not convert parameter vector into image and vice versa.
        But visualization will still be in 2D image format.

    """
    def __init__(self, im_shape, order="C", visual_only=False):
        self._im_shape = im_shape
        self._par_shape = (reduce(operator.mul, im_shape), )
        self.order = order
        self.visual_only = visual_only

        # If visual only, we have same fun_shape as par_shape
        if visual_only:
            self._fun_shape = self._par_shape
        else: # else we have image shape
            self._fun_shape = self._im_shape

    @property
    def fun_shape(self):
        return self._fun_shape

    @property
    def par_shape(self):
        return self._par_shape
    
    @property
    def funvec_shape(self):
        return self.par_shape

    def par2fun(self, pars):
        # If geometry is only used for visualization, do nothing
        if self.visual_only: return pars
        # Else, convert parameter vector into image
        return self._vector_to_image(pars)

    def fun2par(self, funvals):
        # If geometry is only used for visualization, do nothing
        if self.visual_only: return funvals
        # Else, convert image into parameter vector
        return funvals.ravel(order=self.order) #Maybe use reshape((self.dim,), order=self.order)

    def vec2fun(self, funvec):
        """Maps function vector representation, if available, to function values."""    
        return self.par2fun(funvec)
        
    def fun2vec(self, funvals):
        """Maps function values to a vector representation of the function values, if available."""
        return self.fun2par(funvals)
    
    def _vector_to_image(self, vectors):
        """ Converts a vector or multiple vectors into an image. """
        # Reshape to image (also for multiple parameter vectors). TODO: #327
        image = vectors.reshape(self._im_shape+(-1,), order=self.order) 
        #Squeeze to return single image if only one parameter vector was given
        image = image.squeeze(axis=2) if image.shape[2] == 1 else image
        return image
    
    def _plot(self, values, **kwargs):
        # If only visual, we must convert value to image ourselves
        if self.visual_only:
            values = self._vector_to_image(values)

        kwargs.setdefault('cmap', kwargs.get('cmap', "gray"))

        values = self._process_values(values)
        subplot_ids = self._create_subplot_list(values.shape[-1])
        ims = []
        for rows, cols, subplot_id in subplot_ids:
            plt.subplot(rows, cols, subplot_id)
            ims.append(plt.imshow(values[...,subplot_id-1], **kwargs))
        return ims

    def _process_values(self,values):
        if len(values.shape) == 3 or\
             (len(values.shape) == 2 and values.shape[0]== self.par_dim):  
            pass
        else:
            values = values[..., np.newaxis]
        return values

class Discrete(Geometry):
    """ Discrete geometry defined by a list of variables.

    Parameters
    ----------
    variables : list of strings or integer
        If variables is a list, it contains the discrete variables names.
        If variables is an integer, it is the number of discrete variables.
    
    """

    def __init__(self,variables):       
        self.variables = variables

    @property
    def fun_shape(self):
        """The shape of the function space."""
        return (len(self.variables),)

    @property
    def par_shape(self):
        """The shape of the parameter space."""
        return (len(self.variables), )

    def _plot(self,values, **kwargs):

        if ('linestyle' not in kwargs.keys()) and ('ls' not in kwargs.keys()):
            kwargs["linestyle"]  = ''
        
        if ('marker' not in kwargs.keys()):
            kwargs["marker"]  = 'o'

        self._plot_config() 
        return plt.plot(self._ids, values, **kwargs)

    def _plot_envelope(self, lo_values, up_values, **kwargs):
        self._plot_config()
        if 'fmt' in kwargs.keys():
            raise Exception("Argument 'fmt' cannot be passed by the user")

        default = {'color':'dodgerblue', 'fmt':'none' ,'capsize':3, 'capthick':1}
        for key in default:
            if (key not in kwargs.keys()):
                kwargs[key]  = default[key]

        #Convert to 1d numpy array to handle subtraction
        lo_values = np.array(lo_values).flatten()
        up_values = np.array(up_values).flatten()
        
        return plt.errorbar(self._ids, lo_values,
                            yerr=np.vstack(
                                (np.zeros(len(lo_values)), up_values-lo_values)),
                            **kwargs)

    def _plot_config(self):
        # Add at most 10 ticks including the first and last
        n_ticks = min(10, len(self.variables))
        tick_ids = np.linspace(0, len(self.variables)-1, n_ticks, dtype=int)
        plt.xticks(tick_ids, [self.variables[i] for i in tick_ids])

    def fun2par(self,funvals):
        return funvals

class MappedGeometry(_WrappedGeometry):
    """A class that represents a mapped geometry.
    
    This applies a map (callable function) to any cuqi geometry. This will change the par2fun map.
    Additionally an inverse map (imap) can also be defined to allow inverting the function values to parameters redefining the fun2par map.

    Parameters
    -----------
    geometry : cuqi.geometry.Geometry

    map : callable function
        Any callable function representing the map which should be applied after the par2fun of the geometry.

    imap : callable function, Default None
        Any callable function representing the inverse of map.
    """

    def __init__(self,geometry,map,imap=None):
        super().__init__(geometry)
        self.map = map
        self.imap = imap

    def par2fun(self,p):
        return self.map(self.geometry.par2fun(p))

    def fun2par(self,f):
        if self.imap is None:
            raise ValueError("imap is not defined. This is needed for fun2par.")
        return self.geometry.fun2par(self.imap(f))
    
    def fun2vec(self, fun):
        return self.geometry.fun2vec(fun)
    
    def vec2fun(self, funvec):
        return self.geometry.vec2fun(funvec)

    def __repr__(self) -> str:
        return "{}({})".format(self.__class__.__name__,self.geometry.__repr__())

class _DefaultGeometry(Geometry):
    """ Default geometry base class. """
    pass


class _DefaultGeometry1D(Continuous1D, _DefaultGeometry):
    """ Default 1D geometry based on Continuous1D. """
    def __init__(self, grid=None, axis_labels=None):
        super().__init__(grid, axis_labels)

    def __eq__(self, obj):
        if not isinstance(obj, (self.__class__,Continuous1D)): return False
        return self._all_values_equal(obj)
    
class _DefaultGeometry2D(Image2D, _DefaultGeometry):
    """ Default 2D geometry based on Image2D. """
    def __init__(self, im_shape=None, visual_only=False):
        super().__init__(im_shape, visual_only=visual_only)

    def __eq__(self, obj):
        if not isinstance(obj, (self.__class__,Image2D)): return False
        return self._all_values_equal(obj)

# class DiscreteField(Discrete):
#     def __init__(self, grid, cov_func, mean, std, trunc_term=100, axis_labels=['x']):
#         super().__init__(grid, axis_labels)

#         self.N = len(self.grid)
#         self.mean = mean
#         self.std = std
#         XX, YY = np.meshgrid(self.grid, self.grid, indexing='ij')
#         self.Sigma = cov_func(XX, YY)
#         self.L = np.linalg.chol(self.Sigma)

#     def par2fun(self, p):
#         return self.mean + self.L.T@p
    

class KLExpansion(Continuous1D):
    """
    Class representation of the random field in the sine basis


    .. math::
        f = \sum_{i=0}^{N-2} \\left(\\frac{1}{(i+1)^\\gamma\\tau}\\right)  p_i \\, \\text{sin}\\left(\\frac{\\pi}{N}(i+1)(K+\\frac{1}{2})\\right) 
        
        + \\frac{(-1)^K}{2}\\left(\\frac{1}{N^\\gamma\\tau}\\right)  p_{N-1}

    where:

    .. math::
        \\gamma = \\text{decay_rate},

    .. math::
        \\tau = \\text{normalizer},

    :math:`K=\\{0, 1, 2, 3, ..., N-1\\}`, :math:`N` is the number of nodes in the grid, and :math:`p_i` are the expansion coefficients. 

    The above transformation is the inverse of DST-II (see https://en.wikipedia.org/wiki/Discrete_sine_transform)

    Parameters
    -----------
    grid : array-like or None
        One dimensional regular grid on which the random field is defined. If
        grid is None, the user will need to set the grid attribute after
        initialization to be able to fully use the KLExpansion functionalities.

    decay_rate : float, default 2.5
        The decay rate of the basis functions.

    normalizer : float, default 1.0
        A factor of the basis functions shown in the formula above.

    num_modes : int, default None
        Number of expansion modes to use in the KL expansion. If `num_modes` is 
        None or larger than the number of grid points, all modes will be used.

    """

    # init function defining parameters for the KL expansion
    def __init__(self, grid=None, decay_rate=2.5, normalizer=12.0, num_modes=None, axis_labels=None, **kwargs):

        super().__init__(grid, axis_labels, **kwargs)

        self._decay_rate = decay_rate  # decay rate of KL
        self._normalizer = normalizer  # normalizer factor
        self._num_modes = num_modes # number of modes 
        self._coefs = None
        self._coefs_inverse = None

    @property
    def par_shape(self):
        """The shape of the parameter space"""
        return (self.num_modes, )

    @property
    def decay_rate(self):
        return self._decay_rate

    @property
    def normalizer(self):
        return self._normalizer

    @property
    def coefs(self):
        """Computes the deterministic part of the expansion coefficients.
        Returns None if self.num_modes is 0."""
        # Return None if self.num_modes is 0,
        # that is when num_modes is not provided and
        # the grid is None (fun_dim is None).
        if self.num_modes == 0:
            return None

        # If the coefficients are not computed, compute them.
        if self._coefs is None or len(self._coefs) != self.num_modes:
            eigvals = np.array(range(1, self.par_dim+1))  # KL eigvals
            self._coefs = np.diag(1/np.float_power(eigvals, self.decay_rate))

        # Return the coefficients.
        return self._coefs
    
    @property
    def coefs_inverse(self):
        """Computes the inverse of the coefficients diagonal matrix."""
        
        # If the matrix is not computed, compute it.
        if self._coefs_inverse is None or\
            len(self._coefs_inverse) != self.num_modes:
            self._coefs_inverse = np.diag(np.float_power(np.diag(self.coefs), -1))

        # Return the inverse coefficients matrix.
        return self._coefs_inverse

    @property
    def num_modes(self):
        """Number of expansion modes to use in the KL expansion."""
        # If grid is not set, interpret grid dimension as 0.
        grid_dim = self.fun_dim if self.fun_dim is not None else 0

        # If num_modes is not provided or larger than the number of grid points,
        # all modes will be used.
        if self._num_modes is None or self._num_modes > grid_dim:
            return grid_dim
        else:
            return self._num_modes

    # computes the real function out of expansion coefs
    def par2fun(self, p):

        # Reshape the parameter vector
        p = self._reshape_par2fun_input(p)

        modes = self.coefs@p/self.normalizer

        # pad the remaining modes with zeros (for single or multiple parameter
        # parameter vectors)
        modes = np.pad(modes, ((0, self.fun_dim-self.par_dim), (0, 0)),
                       'constant', constant_values=0)

        real = idst(modes.T).T/2

        # squeeze to return single function value if only one parameter vector
        # was given
        return real.squeeze()

    def fun2par(self, funvals):
        """The function to parameter map used to map function values back to
        parameters. In this class (the `KLExpansion`), `fun2par` projects the
        function on the KL expansion coefficient space. Hence this is not
        always the inverse of `par2fun` but it is the closest estimation of the
        function on the KL expansion coefficient space."""

        # Reshape the function values
        funvals = self._reshape_fun2par_input(funvals)

        warnings.warn(
            f"fun2par for {self.__class__} is a projection on "
            + "the KL expansion coefficients space where only "
            + f"the first self.num_modes={self.num_modes} "
            + "coefficients are returned. "
            + _disable_warning_msg("cuqi.geometry")
            + "\n"
        )

        # par2fun scales the input parameters then applies the inverse
        # discrete sine transform of type 2 (IDST-II). Here we apply the
        # inverse of the par2fun map and truncate the expansion
        # coefficients to the number of modes.
        # This includes applying the discrete sine transform of type 2 (DST-II)
        # to the function values to obtain the expansion coefficients.
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.fftpack.dst.html

        # Note that here we have a scaling by 2*self.fun_dim that does
        # not correspond to the scaling in par2fun. This is needed
        # because it is not accounted for in the scipy.fftpack implementation.
        # However, if we use, for example, scipy.fft instead of scipy.fftpack,
        # then this scaling is not needed.

        # Transform (single or multiple functions) to expansion coefficients
        p_temp = dst(funvals.T*2).T[:self.par_dim,:]
        p = self.coefs_inverse@p_temp*self.normalizer/(2*self.fun_dim)
        
        # squeeze to return single parameter vector if only one function value
        # was given
        return p.squeeze()


class KLExpansion_Full(Continuous1D):
    '''
    Class representation of the random field in the sine basis



    .. math::
        f = \\frac{\\text{std}^2}{\\pi}\sum_{i=0}^{N-2} \\left(\\frac{\\tau^\\gamma}{(\\tau+i^2)^\\gamma}\\right)  p_i \\, \\text{sin}\\left(\\frac{\\pi}{N}(i+1)(K+\\frac{1}{2})\\right) 
        
        + \\frac{\\text{std}^2}{\\pi}\\frac{(-1)^K}{2}\\left(\\frac{\\tau^\\gamma}{\\left(\\tau+(N-1)^2\\right)^\\gamma}\\right) p_{N-1}

    where:
    
    .. math::
        \\tau = \\frac{1}{\\text{cor_len}^2},

    .. math::
        \\gamma = \\text{nu}+1,

    :math:`K=\\{0, 1, 2, 3, ..., N-1\\}`, :math:`N` is the number of nodes in the grid, and :math:`p_i` are the expansion coefficients. 

    The above transformation is the inverse of DST-II (see https://en.wikipedia.org/wiki/Discrete_sine_transform)

    Parameters
    -----------
    grid : array-like
        One dimensional regular grid on which the random field is defined.

    cor_len : float, default 1.0
        The correlation length of the random field.

    nu : float, default 2.5
        Smoothness parameter of the random field.

    std : float, default 1.0
        Standard deviation of the random field.
    '''
    
    # init function defining parameters for the KL expansion
    def __init__(self, grid, std=1.0, cor_len=0.2, nu=3.0, axis_labels=None, **kwargs):

        super().__init__(grid, axis_labels, **kwargs)
 
        tau2 = 1./cor_len/cor_len
        gamma = nu+1.
        self.var = std**2

        modes = np.arange(0,self.par_dim)

        self._coefs =  np.float_power( tau2,gamma ) * np.float_power(tau2+modes**2,-gamma)

    @property
    def coefs(self):
        return self._coefs

    # computes the real function out of expansion coefs
    def par2fun(self,p):
        freq = np.zeros(self.par_dim)
        m = len(p)
        freq[:m] = p
        temp = freq*self.coefs
        real = idst(temp)/2/np.pi
        return self.var*real
    
    def fun2par(self,funvals):
        """The function to parameter map used to map function values back to parameters, if available."""
        raise NotImplementedError("fun2par not implemented. ")


class CustomKL(Continuous1D):
    """
    A class representation of a random field in which a truncated KL expansion is computed from a given covariance function.
    
    Parameters
    -----------
    grid : array-like
        One dimensional grid on which the random field is defined.
    
    cov_func : callable
        A covariance function that takes two variables and returns the covariance between them.

    mean : float, default 0.0
        The mean of the random field.

    std : float, default 1.0
        The standard deviation of the random field.
        
    trunc_term : int, default 20% of the number of grid points
        The number of terms to truncate the KL expansion at.
    """
    def __init__(self, grid, mean=0, std=1.0, cov_func=None, trunc_term=None, axis_labels=None, **kwargs):
        super().__init__(grid, axis_labels, **kwargs)

        if trunc_term is None:
            trunc_term = int(len(grid)*0.2)
        self._trunc_term = trunc_term 
        if cov_func is None:
            # Identity covariance function
            cov_func = lambda x,y: 1.0 if np.isclose(x,y,rtol=1e-10) else 0.0
        
        #self.N = len(self.grid)
        self._mean = mean
        self._std = std
        self._compute_eigpairs( grid, cov_func, std, trunc_term, int(2*self.par_dim) )

    @property
    def mean(self):
        return self._mean

    @property
    def std(self):
        return self._std

    @property
    def trunc_term(self):
        return self._trunc_term

    @property
    def eigval(self):
        return self._eigval

    @property
    def eigvec(self):
        return self._eigvec

    @property
    def par_shape(self):
        return (self.trunc_term,)

    def par2fun(self, p):
        return self.mean + ((self.eigvec@np.diag(np.sqrt(self.eigval))) @ p)
    
    def fun2par(self,funvals):
        """The function to parameter map used to map function values back to parameters, if available."""
        raise NotImplementedError("fun2par not implemented. ")
    

    def _compute_eigpairs(self, xnod, C_nu, sigma, M, N_GL):
        # xnod: points at which the field is realized (from geometry PDE)
        # C_nu: lambda function with the covariance kernel (no variance multiplied)
        # sigma: standard deviation of the field
        # M: dimension of the field (KL truncation)
        # N_GL: gauss-legendre points for the integration
        ### return eigenpairs of an arbitrary 1D correlation kernel

        # domain data    
        n = xnod.size
        a = (xnod[-1] - xnod[0])/2     # scale
        
        # compute the Gauss-Legendre abscissas and weights
        #xi, w   = quad.Gauss_Legendre(N_GL) 
        xi, w = np.polynomial.legendre.leggauss(N_GL)

        # transform nodes and weights to [0, L]
        xi_s = a*xi + a
        w_s = a*w
        
        # compute diagonal matrix 
        D = sparse.spdiags(np.sqrt(w_s), 0, N_GL, N_GL).toarray()
        S1 = matlib.repmat(np.sqrt(w_s).reshape(1, N_GL), N_GL, 1)
        S2 = matlib.repmat(np.sqrt(w_s).reshape(N_GL, 1), 1, N_GL)
        S = S1 * S2
        
        # compute covariance matrix 
        Sigma_nu = np.zeros((N_GL, N_GL))
        for i in range(N_GL):
            for j in range(N_GL):
                if i != j:
                    Sigma_nu[i,j] = C_nu(xi_s[i], xi_s[j])
                else:
                    Sigma_nu[i,j] = sigma**2   # correct the diagonal term
                      
        # solve the eigenvalue problem
        A = Sigma_nu * S                             # D_sqrt*Sigma_nu*D_sqrt
        L, h = np.linalg.eig(A)                         # solve eigenvalue problem
        # L, h = sp.sparse.linalg.eigsh(A, M, which='LM')   # np.linalg.eig(A)         
        idx = np.argsort(-np.real(L))                  # index sorting descending
        
        # order the results
        eigval = np.real(L[idx])
        h = h[:,idx]
        
        # take the M values
        eigval = eigval[:M]
        h = np.real(h[:,:M])
        
        # replace for the actual eigenvectors
        phi = np.linalg.solve(D, h)
        
        # Nystrom's interpolation formula
        # recompute covariance matrix on partition nodes and quadrature nodes
        Sigma_nu = np.zeros((n, N_GL))
        for i in range(n):
            for j in range(N_GL):
                Sigma_nu[i,j] = C_nu(xnod[i], xi_s[j])
                          
        M1 = Sigma_nu * np.matlib.repmat(w_s.reshape(N_GL,1), 1, n).T
        M2 = phi @ np.diag(1/eigval)
        eigvec = M1 @ M2

        # normalize eigenvectors (not necessary) integrate to 1
        #norm_fact = np.zeros((M,1))
        #for i in range(M):
        #    norm_fact[i] = np.sqrt(np.trapz(eigvec[:,i]**2, xnod))
        #eigvec = eigvec/np.matlib.repmat(norm_fact, 1, n)             
        self._eigval = eigval
        self._eigvec = eigvec


class StepExpansion(Continuous1D):
    '''
    Class representation of step functions (piecewise constant functions) with `n_steps` 
    equidistant steps on the interval [x0, xn], with both endpoints included.
    The function `par2fun` maps the parameters `p` (which are the step magnitudes) to the 
    corresponding step function evaluated on the spatial grid (`grid`) of nodes x0, x1, ...xn.
    
    For example, if `n_steps` is 3 and `grid` is a uniform grid on [0, L] with nodes x0=0, x1=0.1L, ..., xn=L, 
    then the resulting function evaluated on the grid will be p[0], if x<=L/3, p[1], if L/3<x<=2L/3, 
    p[2], if 2L/3<x<=L.
    
    Parameters
    -----------
    grid: ndarray
        | Regular grid points for the step expansion to be evaluated at. The number of grid points should be equal to or larger than `n_steps`. The latter setting can be useful, for example, when using the StepExpansion geometry as a domain geometry for a cuqipy :class:`Model` that expects the input to be interpolated on a (possibly fine) grid (`grid`). The interval endpoints, [x0, xn], should be included in the grid.

    n_steps: int
        | Number of equidistant steps.

    fun2par_projection: str, default 'mean'
        | Projection of the step function (evaluated on the grid) on the parameter space. The supported projections are 
        | 'mean': the parameter p[i] value will be the average of the function values at the nodes that falls in the interval  I=(x0+i*L/n_steps, x0+(i+1)*L/n_steps].
        | 'max': the parameter p[i] value will be the maximum of the function values in I.
        | 'min': the parameter p[i] value will be the minimum of the function values in I.
        
    kwargs: keyword arguments
        | keyword arguments are passed to the initializer of :class:`~cuqi.geometry.Continuous1D`
    '''
    def __init__(self, grid, n_steps=3, fun2par_projection='mean', **kwargs):

        super().__init__(grid, **kwargs)
        self._n_steps = n_steps
        self._check_grid_setup()
        self._fun2par_projection = fun2par_projection
        L = self.grid[-1]-self.grid[0]
        x0 = self.grid[0]

        self._indices = []
        for i in range(self._n_steps):
            start = x0 + i*L/self._n_steps
            end = x0 + (i+1)*L/self._n_steps
            # Extract indices of the grid points that fall in the ith interval.
            if i ==0:
                interval_indices, =  np.where((self.grid>=start)&(self.grid<=end))
            else:
                interval_indices, = np.where((self.grid>start)&(self.grid<=end))
            self._indices.append(interval_indices)    

    @property
    def par_shape(self):
        """Shape of the parameter space."""
        return (self._n_steps,)

    @property
    def n_steps(self):
        """Number of equidistant steps."""
        return self._n_steps

    def par2fun(self, p):

        # Reshape the parameter vector
        p = self._reshape_par2fun_input(p)

        # Extended fun_shape to include multiple functions.
        ext_fun_shape = self.fun_shape + (p.shape[-1],)

        # Initialize fun to zeros.
        fun = np.zeros(ext_fun_shape)

        # Fill fun with the step function values.
        for i in range(self._n_steps):
            fun[self._indices[i],:] = p[i,:]

        # Squeeze to return single evaluated function if only one parameter
        # vector was given.
        return fun.squeeze()
 

    def fun2par(self,f):

        # Reshape the function values
        f = self._reshape_fun2par_input(f)

        # Extended par_shape to include multiple parameter vectors.
        ext_par_shape = self.par_shape + (f.shape[-1],)
        
        # Initialize par to zeros.
        par = np.zeros(ext_par_shape)

        # Fill par with the projection of the function values.
        for i in range(self._n_steps):
            if self._fun2par_projection.lower() == 'mean':
                projection_method = np.mean
            elif self._fun2par_projection.lower() == 'max':
                projection_method = np.max
            elif self._fun2par_projection.lower() == 'min':
                projection_method = np.min
            else:
                raise ValueError("Invalid projection option.")
            
            # Apply projection method to the function values in the ith
            # interval.
            par[i,:] = projection_method(f[self._indices[i],:], axis=0)

        # Squeeze to return single parameter vector if only one function value
        # was given.
        return par.squeeze()
    

    def _check_grid_setup(self):
        
        # The grid size is greater than or equal to the number of steps.
        if self._n_steps > np.size(self.grid):
            raise ValueError("n_steps must be smaller than the number of grid points")
        
        # Ensure the grid is equally spaced
        if not np.allclose(np.diff(self.grid), self.grid[1]-self.grid[0]):
            raise ValueError("The grid must be an equally spaced grid (regular).")



import scipy as sp
import numpy as np
import cuqi
from cuqi.distribution import Normal
from cuqi.solver import CGLS
from cuqi.sampler import Sampler


class UGLA(Sampler):
    """ Unadjusted (Gaussian) Laplace Approximation sampler
    
    Samples an approximate posterior where the prior is approximated
    by a Gaussian distribution. The likelihood must be Gaussian.

    Currently only works for LMRF priors.

    The inner solver is Conjugate Gradient Least Squares (CGLS) solver.

    For more details see: Uribe, Felipe, et al. "A hybrid Gibbs sampler for edge-preserving 
    tomographic reconstruction with uncertain view angles." arXiv preprint arXiv:2104.06919 (2021).

    Parameters
    ----------
    target : `cuqi.distribution.Posterior`
        The target posterior distribution to sample.

    x0 : ndarray
        Initial parameters. *Optional*

    maxit : int
        Maximum number of inner iterations for solver when generating one sample.

    tol : float
        Tolerance for inner solver. Will stop before maxit if the inner solvers convergence check reaches tol.

    beta : float
        Smoothing parameter for the Gaussian approximation of the Laplace distribution. Larger beta is easier to sample but is a worse approximation.

    rng : np.random.RandomState
        Random number generator used for sampling. *Optional*

    callback : callable, *Optional*
        If set this function will be called after every sample.
        The signature of the callback function is `callback(sample, sample_index)`,
        where `sample` is the current sample and `sample_index` is the index of the sample.
        An example is shown in demos/demo31_callback.py.

    Returns
    -------
    cuqi.samples.Samples
        Samples from the posterior distribution.

    """

    def __init__(self, target, x0=None, maxit=50, tol=1e-4, beta=1e-5, rng=None, **kwargs):
        
        super().__init__(target, x0=x0, **kwargs)

        # Check target type
        if not isinstance(self.target, cuqi.distribution.Posterior):
            raise ValueError(f"To initialize an object of type {self.__class__}, 'target' need to be of type 'cuqi.distribution.Posterior'.")       

        # Check Affine model
        if not isinstance(self.target.likelihood.model, cuqi.model.AffineModel):
            raise TypeError("Model needs to be affine or linear")

        # Check Gaussian likelihood
        if not hasattr(self.target.likelihood.distribution, "sqrtprec"):
            raise TypeError("Distribution in Likelihood must contain a sqrtprec attribute")

        # Check that prior is LMRF
        if not isinstance(self.target.prior, cuqi.distribution.LMRF):
            raise ValueError('Unadjusted Gaussian Laplace approximation (UGLA) requires LMRF prior')

        # Modify initial guess since Sampler sets it to ones.       
        if x0 is not None:
            self.x0 = x0
        else:
            self.x0 = np.zeros(self.target.prior.dim)
        
        # Store internal parameters
        self.maxit = maxit
        self.tol = tol
        self.beta = beta
        self.rng = rng

    def _sample_adapt(self, Ns, Nb):
        return self._sample(Ns, Nb)

    def _sample(self, Ns, Nb):
        """ Sample from the approximate posterior.

        Parameters
        ----------
        Ns : int
            Number of samples to draw.

        Nb : int
            Number of burn-in samples to discard.

        Returns
        -------
        samples : ndarray
            Samples from the approximate posterior.

        target_eval : ndarray
            Log-likelihood of each sample.

        acc : ndarray
            Acceptance rate of each sample.

        """

        # Extract diff_op from target prior
        D = self.target.prior._diff_op
        n = D.shape[0]

        # Gaussian approximation of LMRF prior as function of x_k
        def Lk_fun(x_k):
            dd =  1/np.sqrt((D @ x_k)**2 + self.beta*np.ones(n))
            W = sp.sparse.diags(dd)
            return W.sqrt() @ D

        # Now prepare "LinearRTO" type sampler. TODO: Use LinearRTO for this instead
        self._shift = 0

        # Pre-computations
        self._model = self.target.likelihood.model   
        self._data = self.target.likelihood.data - self.target.model._shift
        self._m = len(self._data)
        self._L1 = self.target.likelihood.distribution.sqrtprec

        # If prior location is scalar, repeat it to match dimensions
        if len(self.target.prior.location) == 1:
            self._priorloc = np.repeat(self.target.prior.location, self.dim)
        else:
            self._priorloc = self.target.prior.location

        # Initial Laplace approx
        self._L2 = Lk_fun(self.x0)
        self._L2mu = self._L2@self._priorloc
        self._b_tild = np.hstack([self._L1@self._data, self._L2mu]) 
        
        #self.n = len(self.x0)
        
        # Least squares form
        def M(x, flag):
            if flag == 1:
                out1 = self._L1 @ self._model._forward_func_no_shift(x) # Use forward function which excludes shift
                out2 = np.sqrt(1/self.target.prior.scale)*(self._L2 @ x)
                out  = np.hstack([out1, out2])
            elif flag == 2:
                idx = int(self._m)
                out1 = self._model._adjoint_func_no_shift(self._L1.T@x[:idx])
                out2 = np.sqrt(1/self.target.prior.scale)*(self._L2.T @ x[idx:])
                out  = out1 + out2                
            return out 
        
        # Initialize samples
        N = Ns+Nb   # number of simulations        
        samples = np.empty((self.target.dim, N))
                     
        # initial state   
        samples[:, 0] = self.x0
        for s in range(N-1):

            # Update Laplace approximation
            self._L2 = Lk_fun(samples[:, s])
            self._L2mu = self._L2@self._priorloc
            self._b_tild = np.hstack([self._L1@self._data, self._L2mu]) 
        
            # Sample from approximate posterior
            e = Normal(mean=np.zeros(len(self._b_tild)), std=1).sample(rng=self.rng)
            y = self._b_tild + e # Perturb data
            sim = CGLS(M, y, samples[:, s], self.maxit, self.tol, self._shift)            
            samples[:, s+1], _ = sim.solve()

            self._print_progress(s+2,N) #s+2 is the sample number, s+1 is index assuming x0 is the first sample
            self._call_callback(samples[:, s+1], s+1)

        # remove burn-in
        samples = samples[:, Nb:]
        
        return samples, None, None



class _OrderedSet:
    """A set (i.e. unique elements) that keeps its elements in the order they were added.

    This is a minimal implementation of an ordered set, using a dictionary for storage.
    """
    
    def __init__(self, iterable=None):
        """Initialize the OrderedSet. 

        If an iterable is provided, add all its elements to the set.
        """
        self.dict = dict.fromkeys(iterable if iterable else [])

    def add(self, item):
        """Add an item to the set. 

        If the item is already in the set, it does nothing.
        Otherwise, the item is stored as a key in the dictionary, with None as its value.
        """
        self.dict[item] = None

    def remove(self, item):
        """Remove an item from the set. 

        If the item is not in the set, it raises a KeyError.
        """
        del self.dict[item]

    def __contains__(self, item):
        """Check if an item is in the set. 

        This is equivalent to checking if the item is a key in the dictionary.
        """
        return item in self.dict

    def __iter__(self):
        """Return an iterator over the set. 

        This iterates over the keys in the dictionary.
        """
        return iter(self.dict)

    def __len__(self):
        """Return the number of items in the set."""
        return len(self.dict)

    def extend(self, other):
        """Extend the set with the items in another set.

        Raises a TypeError if the other object is not an _OrderedSet.
        """
        if not isinstance(other, _OrderedSet):
            raise TypeError("unsupported operand type(s) for extend: '_OrderedSet' and '{}'".format(type(other).__name__))
        for item in other:
            self.add(item)

    def replace(self, old_item, new_item):
        """Replace old_item with new_item at the same position, preserving order."""
        if old_item not in self.dict:
            raise KeyError(f"{old_item} not in set")
        
        items = list(self.dict.keys())  # Preserve order
        index = items.index(old_item)  # Find position
        items[index] = new_item  # Replace at the same position

        # Reconstruct the ordered set with the new item in place
        self.dict = dict.fromkeys(items)

    def __or__(self, other):
        """Return a new set that is the union of this set and another set.

        Raises a TypeError if the other object is not an _OrderedSet.
        """
        if not isinstance(other, _OrderedSet):
            raise TypeError("unsupported operand type(s) for |: '_OrderedSet' and '{}'".format(type(other).__name__))
        new_set = _OrderedSet(self.dict.keys())
        new_set.extend(other)
        return new_set

    def __repr__(self):
        """Return a string representation of the set."""
        return "_OrderedSet({})".format(list(self.dict.keys()))



from math import isnan
import cuqi
import numpy as np
import scipy as sp
import scipy.stats as scipy_stats
import scipy.sparse as sps
import numpy.linalg as nplinalg

from pytest import approx
import pytest

def test_Normal_mean_standard():
    assert cuqi.distribution.Normal(0,1).mean == approx(0.0)

def test_Normal_pdf_mean():
    pX = cuqi.distribution.Normal(0.1,1)
    assert pX.pdf(0.1) == approx(1.0/np.sqrt(2.0*np.pi))

@pytest.mark.parametrize("mean,var,expected",[
                            (2,3.5,[[8.17418321], [3.40055023]]),
                            (3.141592653589793,2.6457513110645907,[[7.80883646],[4.20030911]]),
                            (-1e-09, 1000000.0,[[1764052.34596766],[400157.20836722]]),
                            (1.7724538509055159, 0, [[1.77245385],[1.77245385]])
                        ])
def test_Normal_sample_regression(mean,var,expected):
    rng = np.random.RandomState(0) #Replaces legacy method: np.random.seed(0)
    samples = cuqi.distribution.Normal(mean,var).sample(2,rng=rng)
    target = np.array(expected).T
    assert np.allclose( samples.samples, target)

@pytest.mark.parametrize("mean,std,points,expected",[
                            (0,1,[-1,0,1],[1,0,-1]),
   (np.array([0,0]),np.array([1,1]),[[-1,0],[0,0],[0,-1]], [[1,0],[0,0],[0,1]])])
def test_Normal_gradient(mean,std,points,expected):
    p = cuqi.distribution.Normal(mean,std)
    for point, grad in zip(points, expected):
        assert np.allclose(p.gradient(point), grad)

@pytest.mark.parametrize("mean,std,low,high,points",[(0.0, 
                                               1.0,
                                               -1.0,
                                               1.0,
                                               [-1.5, -0.5, 0.5, 1.5]),
                                              (np.array([0.0, 0.0]), 
                                               np.array([1.0, 1.0]),
                                               np.array([-1.0, -1.0]),
                                               np.array([1.0, 1.0]),
                                               [np.array([-0.5, 0.0]),
                                                np.array([0.5, 0.0]),
                                                np.array([-2.0, 0.0]),
                                                np.array([2.0, 0.0])])])
def test_TruncatedNormal_logpdf(mean,std,low,high,points):
    x_trun = cuqi.distribution.TruncatedNormal(mean,std,low=low,high=high)
    x = cuqi.distribution.Normal(mean,std)
    for point in points:
        if np.all(point >= low) and np.all(point <= high):
            assert x_trun.logpdf(point) == approx(x.logpdf(point))
        else:
            assert np.isneginf(x_trun.logpdf(point))

@pytest.mark.parametrize("mean,std,low,high,points",[(0.0, 
                                               1.0,
                                               -1.0,
                                               1.0,
                                               [-1.5, -0.5, 0.5, 1.5]),
                                              (np.array([0.0, 0.0]), 
                                               np.array([1.0, 1.0]),
                                               np.array([-1.0, -1.0]),
                                               np.array([1.0, 1.0]),
                                               [np.array([-0.5, 0.0]),
                                                np.array([0.5, 0.0]),
                                                np.array([-2.0, 0.0]),
                                                np.array([2.0, 0.0])])])
def test_TruncatedNormal_gradient(mean,std,low,high,points):
    x_trun = cuqi.distribution.TruncatedNormal(mean,std,low=low,high=high)
    x = cuqi.distribution.Normal(mean,std)
    for point in points:
        if np.all(point >= low) and np.all(point <= high):
            assert np.all(x_trun.gradient(point) == approx(x.gradient(point)))
        else:
            assert np.all(np.isnan(x_trun.gradient(point)))

@pytest.mark.parametrize("mean,std,low,high",[(0.0, 
                                               1.0,
                                               -1.0,
                                               1.0),
                                              (np.array([0.0, 0.0]), 
                                               np.array([1.0, 1.0]),
                                               np.array([-1.0, -1.0]),
                                               np.array([1.0, 1.0]))])
def test_TruncatedNormal_sampling(mean,std,low,high):
    x = cuqi.distribution.TruncatedNormal(mean,std,low=low,high=high)
    samples = x.sample(10000).samples
    for i in range(samples.shape[1]):
        sample = samples[:,i]
        assert np.all(sample >= low) and np.all(sample <= high)

def test_Gaussian_mean():
    mean = np.array([0, 0])
    std = np.array([1, 1])
    R = np.array([[1, -0.7], [-0.7, 1]])
    cov = np.diag(std) @ (R @ np.diag(std))
    pX_1 = cuqi.distribution.Gaussian(mean, cov)
    assert np.allclose(pX_1.mean, np.array([0, 0]) ) 

def test_Gaussian_cov():
    mean = np.array([0, 0])
    std = np.array([1.3, 2])
    R = np.array([[1, -0.7], [-0.7, 1]])
    D = np.diag(std)
    S = D @ R @ D
    pX_1 = cuqi.distribution.Gaussian(mean, S)
    assert np.allclose(pX_1.cov, S)
    
def test_Gaussian_multiple():
    pX_1 = cuqi.distribution.Gaussian(np.zeros(2), np.array([[1.5, -.5],[-.5, 1]]))
    #
    X, Y = np.meshgrid(np.linspace(-4, 4, 200), np.linspace(-4, 4, 200))
    Xf, Yf = X.flatten(), Y.flatten()
    pts = np.vstack([Xf, Yf]).T   # pts is (m*n, d)
    npoints = pts.shape[0]
    Z = pX_1.pdf(pts)
    assert npoints == Z.shape[0]

@pytest.mark.xfail(reason="Expected to fail after fixing Gaussian sample. Regression needs to be updated")
@pytest.mark.parametrize("mean,std,R,expected",[
                        (([0, 0]),
                        ([1, 1]),
                        ([[ 1.0 , -0.7],
                          [-0.7,  1. ]]),
                        ([[ 0.30471644,  0.52294788,  0.3204432 ,  0.82791771,  0.88232622],
                          [-2.52738159,  0.50701152, -1.04189629, -2.16116453, -1.34325028]])),

                        (([-3.14159265,  2.23606798]),
                        ([ 3.14159265, 50.        ]),
                        ([[ 1.0   ,  -0.001],
                          [-0.001,  1.0   ]]),
                        ([[  2.40014477,  -1.88427406,  -0.06682814,   3.89835695,  2.72559222],
                          [-46.63338991,  49.73922674,  -5.33487942,  -2.93194249, 22.76010272]])),

                        (([23.        ,  0.0        ,  3.14159265]),
                        ([3.0        , 1.41421356, 3.14159265]),
                        ([[ 1.0, 0.9,  0.3],
                          [0.9,  1. , -0.5],
                          [0.3, -0.5,  1.0 ]]),
                        ([[23.7305748 , 22.58046919, 23.19981226, 23.43140839, 23.1244483 ],
                          [ 2.83554239,  3.92371224,  3.06634238,  4.00865492,  4.22990048],
                          [ 0.39008306,  4.56363292,  2.81075259, -1.94308617, -0.87372503]]))
                        ])
def test_Gaussian_sample_regression(mean,std,R,expected):
    rng = np.random.RandomState(0)
    std = np.array(std)
    R = np.array(R)
    cov = np.diag(std) @ (R @ np.diag(std))
    pX_1 = cuqi.distribution.Gaussian(np.array(mean), cov)
    samples = pX_1.sample(5,rng=rng).samples
    assert np.allclose( samples, np.array(expected))

@pytest.mark.parametrize("seed",[3,4,5],ids=["seed3","seed4","seed5"])
@pytest.mark.parametrize("mean,var",[(2,3),(np.pi,0),(np.sqrt(5),1)]) 
def test_Normal_rng(mean,var,seed):
    np.random.seed(seed)
    rng = np.random.RandomState(seed)
    assert np.allclose(cuqi.distribution.Normal(mean,var).sample(10).samples,cuqi.distribution.Normal(mean,var).sample(10,rng=rng).samples)

@pytest.mark.parametrize("mean,std,R",[
                        (([0, 0]),
                        ([1, 1]),
                        ([[ 1. , -0.7],
                          [-0.7,  1. ]])),
                        (([-3.14159265,  2.23606798]),
                        ([ 3.14159265, 50.        ]),
                        ([[ 1.   ,  -0.001],
                          [-0.001,  1.   ]])),
                        ])
def test_Gaussian_rng(mean,std,R):
    np.random.seed(3)
    rng = np.random.RandomState(3)
    std = np.array(std)
    R = np.array(R)
    cov = np.diag(std) @ (R @ np.diag(std))
    assert np.allclose(cuqi.distribution.Gaussian(mean,cov).sample(10).samples,cuqi.distribution.Gaussian(mean,cov).sample(10,rng=rng).samples)

@pytest.mark.parametrize("dist",[cuqi.distribution.GMRF(np.ones(128),35,'zero'),cuqi.distribution.GMRF(np.ones(128),35,'periodic'),cuqi.distribution.GMRF(np.ones(128),35,'neumann')])
def test_GMRF_rng(dist):
    np.random.seed(3)
    rng = np.random.RandomState(3)
    assert np.allclose(dist.sample(10).samples,dist.sample(10,rng=rng).samples)

@pytest.mark.parametrize( \
  "low,high,toeval,expected",[ \
    (-2.0, 3.0, 1.0, np.log(0.2)), \
    (-2.0, 3.0, 3.5, -np.inf), \
    (np.array([1.0, 0.5]), np.array([2.0, 2.5]), np.array([1, 0.5]), np.log(0.5)), \
    (np.array([1.0, 0.5]), np.array([2.0, 2.5]), np.array([0.0, 0.5]), -np.inf), \
    (np.array([1.0, 0.5]), np.array([2.0, 2.5]), np.array([0.0, 0.0]), -np.inf), \
    (np.array([1.0, 0.5]), np.array([2.0, 2.5]), np.array([3.0, 0.5]), -np.inf), \
    (np.array([1.0, 0.5]), np.array([2.0, 2.5]), np.array([3.0, 3.0]), -np.inf), \
    (np.array([1.0, 2.0, 3.0]), np.array([3.0, 4.0, 5.0]), np.array([3.0, 4.0, 3.0]), np.log(0.125)), \
    (np.array([1.0, 2.0, 3.0]), np.array([3.0, 4.0, 5.0]), np.array([0.0, 5.0, 4.0]), -np.inf) \
  ])
def test_Uniform_logpdf(low, high, toeval, expected):
    UD = cuqi.distribution.Uniform(low, high)
    assert np.allclose(UD.logpdf(toeval), expected) 

@pytest.mark.parametrize("low,high,expected",[(np.array([1, .5]), 
                                               np.array([2, 2.5]),
                                               np.array([[1.5507979 , 1.29090474, 1.89294695],
                                               [1.91629565, 1.52165521, 2.29258618]])),
                                              (1,2,
                                               np.array([[1.5507979 , 1.70814782, 1.29090474]]))])
def test_Uniform_sample(low, high, expected):
    rng = np.random.RandomState(3)
    UD = cuqi.distribution.Uniform(low, high)
    cuqi_samples = UD.sample(3,rng=rng)
    print(cuqi_samples)
    assert np.allclose(cuqi_samples.samples, expected) 

def test_UnboundedUniform_logpdf_and_gradient():
    x1 = cuqi.distribution.UnboundedUniform(geometry=1)
    assert np.allclose(x1.logpdf(np.array([1])), 1.0)
    assert np.allclose(x1.gradient(np.array([0])), np.array([0]))
    x2 = cuqi.distribution.UnboundedUniform(geometry=cuqi.geometry.Discrete(2))
    assert np.allclose(x2.logpdf(np.array([1, 2])), 1.0)
    assert np.allclose(x2.gradient(np.array([0, 0])), np.array([0, 0]))

def test_UnboundedUniform_sample():
    x = cuqi.distribution.UnboundedUniform(geometry=2)
    with pytest.raises(NotImplementedError, match="Cannot sample from UnboundedUniform distribution"):
        x.sample()

@pytest.mark.parametrize("distribution, kwargs",
                         [(cuqi.distribution.Uniform, 
                          {'low':np.array([2, 2.5, 3,5]),
                          'high':np.array([5,7, 7,6])}),
                          (cuqi.distribution.Gaussian, 
                          {'mean':np.array([0, 0, 0, 0]),
                          'sqrtcov':np.array([1, 1, 1, 1])})
                          ])
def test_distribution_contains_geometry(distribution, kwargs):
    rng = np.random.RandomState(3)
    geom = cuqi.geometry.Continuous2D((2,2))
    dist = distribution(**kwargs,geometry = geom)
    cuqi_samples = dist.sample(3,rng=rng)
    assert(dist.dim == geom.par_dim and 
          cuqi_samples.geometry == geom and
          dist.geometry == geom and 
          np.all(geom.grid[0]==np.array([0, 1])) and
          np.all(geom.grid[1]== np.array([0, 1])))

# Compare computed covariance
@pytest.mark.parametrize("seed", [0,3,5,7])
@pytest.mark.parametrize("mean,cov,mean_full,cov_full",[
    ( (0),           (5),            (0),           (5)           ),
    ( (0),           (5*np.ones(3)), (np.zeros(3)), (5*np.eye(3)) ),
    ( (np.zeros(3)), (5),            (np.zeros(3)), (5*np.eye(3)) ),
    ( (np.zeros(3)), (5*np.ones(3)), (np.zeros(3)), (5*np.eye(3)) ),
    ( (0),           (5*np.eye(3)),  (np.zeros(3)), (5*np.eye(3)) ),
    ( (0),           (5*sps.eye(3)), (np.zeros(3)), (5*np.eye(3)) ),
    ( (0), (np.array([[5,-3],[-3,2]])),       (np.zeros(2)), (np.array([[5,-3],[-3,2]])) ),
    #( (0), (sps.csc_matrix([[5,3],[-3,2]])), (np.zeros(2)), (np.array([[5,3],[-3,2]])) ),
])
def test_Gaussian_cov(mean,cov, mean_full, cov_full, seed):
    #Fix seed for reproducibility
    np.random.seed(seed)

    # Define cuqi dist using various means and covs
    prior = cuqi.distribution.Gaussian(mean, cov)

    # Compare logpdf with scipy using full vector+matrix rep
    x0 = 1000*np.random.rand(prior.dim)
    eval1 = prior.logpdf(x0)
    eval2 = sp.stats.multivariate_normal.logpdf(x0, mean_full, cov_full)

    assert np.allclose(eval1,eval2)

    gradeval1 = prior.gradient(x0)
    gradeval2 = sp.optimize.approx_fprime(x0, prior.logpdf, 1e-15)

    assert np.allclose(gradeval1,gradeval2)


def _stats(samples):
    """ Compute meadian, std, and lo95, up95 of samples """
    return np.vstack((np.median(samples, axis=1), np.std(samples, axis=1), np.percentile(samples, 2.5, axis=1), np.percentile(samples, 97.5, axis=1)))

@pytest.mark.parametrize("prec, GMRF_order",[
    (sps.diags([-1, 2, -1], [-1, 0, 1], shape=(5, 5)), 1),
    #(sps.diags([-1, 2, -1], [-1, 0, 1], shape=(128, 128)), 1),
    (sps.diags([1, -4, 6, -4, 1], [-2, -1, 0, 1, 2], shape=(5, 5)), 2),
    #(sps.diags([1, -4, 6, -4, 1], [-2, -1, 0, 1, 2], shape=(128, 128)), 2),
    (sps.eye(5), 0),
])
def test_Gaussians_vs_GMRF(prec, GMRF_order):
    """ Tests the various Gaussians given some common precision matrices related to GMRFs
    
    This tests both logpdf, gradient, and sample methods for sparse and dense matricies as input.
    """

    # Get dimension of precision matrix
    dim = prec.shape[0]

    # Store normal and sparse version of precision matrix
    prec_s = prec
    prec = prec.toarray()

    # Compute covariance from precision
    cov = sp.linalg.inv(prec)
    cov_s = sps.linalg.inv(prec_s)

    # Compute lower triangular Cholesky decomposition of covariance
    sqrtcov = sp.linalg.cholesky(cov, lower=True)
    sqrtcov_s = sp.sparse.csr_matrix(sqrtcov)  

    # Compute upper triangular cholesky decomposition of precision
    sqrtprec = sp.linalg.cholesky(prec)
    sqrtprec_s = sp.sparse.csr_matrix(sqrtprec)    

    # Define Gaussians from all combinations
    X_prec = cuqi.distribution.Gaussian(np.zeros(dim), prec=prec)
    X_cov = cuqi.distribution.Gaussian(np.zeros(dim), cov)
    X_sqrtprec = cuqi.distribution.Gaussian(np.zeros(dim), sqrtprec=sqrtprec)
    X_sqrtcov = cuqi.distribution.Gaussian(np.zeros(dim), sqrtcov=sqrtcov)
    X_GMRF = cuqi.distribution.GMRF(np.zeros(dim), 1, 'zero', order=GMRF_order)

    # logpdfs for full matrix
    x0 = np.random.randn(dim)
    assert np.allclose(X_cov.logpdf(x0), X_GMRF.logpdf(x0))
    assert np.allclose(X_cov.logpdf(x0), X_sqrtprec.logpdf(x0))
    assert np.allclose(X_cov.logpdf(x0), X_prec.logpdf(x0))
    assert np.allclose(X_cov.logpdf(x0), X_sqrtcov.logpdf(x0))

    # gradients for full matrix
    assert np.allclose(X_cov.gradient(x0), X_GMRF.gradient(x0))
    #assert np.allclose(X_cov.gradient(x0), X_sqrtprec.gradient(x0)) #TODO: NotImplementedError
    assert np.allclose(X_cov.gradient(x0), X_prec.gradient(x0))

    # samples (compare statistics)
    Ns = 10000
    s_cov = _stats(X_cov.sample(Ns).samples)
    s_GMRF = _stats(X_GMRF.sample(Ns).samples)
    s_sqrtprec = _stats(X_sqrtprec.sample(Ns).samples)
    s_prec = _stats(X_prec.sample(Ns).samples)
    s_sqrtcov = _stats(X_sqrtcov.sample(Ns).samples)

    # We round to one decimal and allow 10% error in sample statistics.
    # TODO. Better comparrison here..
    # We want to get a precision of e.g. 2 in scientific notation
    # float(np.format_float_scientific(2.367e-01, precision=1) = 0.24
    # float(np.format_float_scientific(9.650e+01, precision=1)) = 96.0
    # its not the same as np.round().
    assert np.allclose(np.round(s_cov, 1), np.round(s_GMRF, 1) , rtol=0.1)
    assert np.allclose(np.round(s_cov, 1), np.round(s_sqrtprec, 1) , rtol=0.1)
    assert np.allclose(np.round(s_cov, 1), np.round(s_prec, 1) , rtol=0.1)
    assert np.allclose(np.round(s_cov, 1), np.round(s_sqrtcov, 1) , rtol=0.1)

    # Now compare sparse versions
    # Check un-normalized logpdfs for sparse precision and covariance
    X_prec_s = cuqi.distribution.Gaussian(np.zeros(dim), prec=prec_s)
    X_cov_s = cuqi.distribution.Gaussian(np.zeros(dim), cov=cov_s)
    X_sqrtprec_s = cuqi.distribution.Gaussian(np.zeros(dim), sqrtprec=sqrtprec_s)
    X_sqrtcov_s = cuqi.distribution.Gaussian(np.zeros(dim), sqrtcov=sqrtcov_s)

    assert np.allclose(X_cov._logupdf(x0), X_cov_s._logupdf(x0))
    assert np.allclose(X_cov._logupdf(x0), X_prec_s._logupdf(x0))
    assert np.allclose(X_cov._logupdf(x0), X_sqrtprec_s._logupdf(x0))
    assert np.allclose(X_cov._logupdf(x0), X_sqrtcov_s._logupdf(x0))

    # Check gradients for sparse precision and covariance
    assert np.allclose(X_cov.gradient(x0), X_cov_s.gradient(x0))
    assert np.allclose(X_cov.gradient(x0), X_prec_s.gradient(x0))
    #assert np.allclose(X_cov.gradient(x0), X_sqrtprec_s.gradient(x0)) # TODO: NotImplementedError
    assert np.allclose(X_cov.gradient(x0), X_sqrtcov_s.gradient(x0))

    # Check samples for sparse precision and covariance
    s_cov_s = _stats(X_cov_s.sample(Ns).samples)
    s_prec_s = _stats(X_prec_s.sample(Ns).samples)
    s_sqrtprec_s = _stats(X_sqrtprec_s.sample(Ns).samples)
    s_sqrtcov_s = _stats(X_sqrtcov_s.sample(Ns).samples)

    assert np.allclose(np.round(s_cov, 1), np.round(s_cov_s, 1) , rtol=0.1)
    assert np.allclose(np.round(s_cov, 1), np.round(s_prec_s, 1) , rtol=0.1)
    assert np.allclose(np.round(s_cov, 1), np.round(s_sqrtprec_s, 1) , rtol=0.1)
    assert np.allclose(np.round(s_cov, 1), np.round(s_sqrtcov_s, 1) , rtol=0.1)
    
    #TODO. Add comparrison of sampling using X_cov.sqrtprec directly. This is what LinearRTO uses.
    # CUQI test problem
    # TP = cuqi.testproblem.Deconvolution1D(dim=n)
    # TP.prior = X_GMRF
    # samples_GMRF = TP._sampleLinearRTO(2000)
    # TP.prior = X_cov
    # samples_cov = TP._sampleLinearRTO(2000)
    # TP.prior = X_sqrtprec
    # samples_sqrtprec = TP._sampleLinearRTO(2000)
    # samples_GMRF.plot_ci()
    # samples_sqrtprec.plot_ci()
    # samples_cov.plot_ci()

def test_InverseGamma_sample():
    a = [1,2]
    location = 0
    scale = 1
    N = 1000
    rng1 = np.random.RandomState(1)
    rng2 = np.random.RandomState(1)
    x = cuqi.distribution.InverseGamma(shape=a, location=location, scale=scale)
    samples1 = x.sample(N, rng=rng1).samples
    samples2 = sp.stats.invgamma.rvs(a=a, loc=location, scale=scale, size=(N,len(a)), random_state=rng2).T

    assert x.dim == len(a)
    assert np.all(np.isclose(samples1, samples2))

@pytest.mark.parametrize("a, location, scale",
                         [([1,2,1], [0,-1, 100], np.array([.1, 1, 20])),
                          (np.array([3,2,1]), (0, 0, 0), 1)])
@pytest.mark.parametrize("x",
                         [(np.array([1, 4, .5])),
                          (np.array([6, 6, 120])),
                          (np.array([1000, 0, -40]))])
@pytest.mark.parametrize("func", [("pdf"),("cdf"),("logpdf"),("gradient")])                        
def test_InverseGamma(a, location, scale, x, func):
    IGD = cuqi.distribution.InverseGamma(a, location=location, scale=scale)

    # PDF formula for InverseGamma
    def my_pdf( x, a, location, scale):
        if np.any(x <= location):
            val = x*0
        else:
            val =  (scale**a)\
            /((x-location)**(a+1)*sp.special.gamma(a))\
            *np.exp(-scale/(x-location))
        return val

    if func == "pdf":
        # The joint PDF of independent random vairables is the product of their individual PDFs.
        print("#########")
        print(IGD.pdf(x))
        print(np.prod(my_pdf(x, IGD.shape, IGD.location, IGD.scale)))

        assert np.all(np.isclose(IGD.pdf(x),np.prod(sp.stats.invgamma.pdf(x, a=a, loc=location, scale=scale)))) and np.all(np.isclose(IGD.pdf(x),np.prod(my_pdf(x, IGD.shape, IGD.location, IGD.scale))))

    elif func == "cdf":
        # The joint CDF of independent random vairables is the product of their individual CDFs.
        assert np.all(np.isclose(IGD.cdf(x),np.prod(sp.stats.invgamma.cdf(x, a=a, loc=location, scale=scale))))

    elif func == "logpdf":
        # The joint PDF of independent random vairables is the product of their individual PDFs (the product is replaced by sum for logpdf).
        assert np.all(np.isclose(IGD.logpdf(x),np.sum(sp.stats.invgamma.logpdf(x, a=a, loc=location, scale=scale)))) and np.all(np.isclose(IGD.logpdf(x),np.sum(np.log(my_pdf(x, IGD.shape, IGD.location, IGD.scale)))))

    elif func == "gradient":
        FD_gradient = cuqi.utilities.approx_gradient(IGD.logpdf, x, epsilon=0.000000001)
        #Assert that the InverseGamma gradient is close to the FD approximation or both gradients are nan.
        assert (np.all(np.isclose(IGD.gradient(x),FD_gradient,rtol=1e-3)) or
          (np.all(np.isnan(FD_gradient)) and np.all(np.isnan(IGD.gradient(x)))))

    else:
        raise ValueError
    
@pytest.mark.parametrize("shape", [1, 2, 3, 5])
@pytest.mark.parametrize("rate", [1e-4, 1e-3, 1e-2, 1e-1, 1, 1e4, 1e5])
@pytest.mark.parametrize("value", [1, 2, 3, 4, 5])
def test_Gamma_pdf(shape, rate, value):
    G = cuqi.distribution.Gamma(shape, rate)
    assert np.isclose(G.pdf(value), scipy_stats.gamma(shape, scale=1/rate).pdf(value))

@pytest.mark.parametrize("shape", [1, 2, 3, 5])
@pytest.mark.parametrize("rate", [1e-4, 1e-3, 1e-2, 1e-1, 1, 1e4, 1e5])
@pytest.mark.parametrize("value", [1, 2, 3, 4, 5])
def test_Gamma_cdf(shape, rate, value):
    G = cuqi.distribution.Gamma(shape, rate)
    assert np.isclose(G.cdf(value), scipy_stats.gamma(shape, scale=1/rate).cdf(value))

@pytest.mark.parametrize("shape", [1, 2, 3, 5])
@pytest.mark.parametrize("rate", [1e-4, 1e-3, 1e-2, 1e-1, 1, 1e4, 1e5])
def test_Gamma_sample(shape, rate):
    rng = np.random.RandomState(3)
    G = cuqi.distribution.Gamma(shape, rate)
    cuqi_samples = G.sample(3, rng=rng)

    rng2 = np.random.RandomState(3)
    np_samples = rng2.gamma(shape=shape, scale=1/rate, size=(3, 1)).T

    assert np.allclose(cuqi_samples.samples, np_samples)

@pytest.mark.parametrize("location", [-1, -2, -3, 0, 1, 2, 3])
@pytest.mark.parametrize("scale", [1e-3, 1e-1, 1e0, 1e1, 1e3])
@pytest.mark.parametrize("value", [1e-3, 1e-1, 1e0, 1e1, 1e3])
def test_Laplace_pdf(location, scale, value):
    LPL = cuqi.distribution.Laplace(location, scale)
    assert np.isclose(LPL.pdf(value), scipy_stats.laplace(location, scale).pdf(value))

@pytest.mark.xfail(reason="Expected to fail after fixing Gaussian sample. Regression needs to be updated")
def test_lognormal_sample():
    rng = np.random.RandomState(3)
    mean = np.array([0, -4])
    std = np.array([1, 14])
    R = np.array([[1, -0.7], [-0.7, 1]])
    LND = cuqi.distribution.Lognormal(mean, std**2*R)
    cuqi_samples = LND.sample(3,rng=rng)
    result = np.array([[5.92800307e+00, 1.54490711e+00, 1.09977286e+00],
                       [7.82563924e-14, 3.68757304e-04, 1.26980745e-04]])
    assert(np.all(np.isclose(cuqi_samples.samples, result)))

@pytest.mark.parametrize("mean,std",[
                        (np.array([0, 0]),
                        np.array([1, 1])),
                        (np.array([-3.14159265,  2.23606798]),
                        np.array([3.14159265, 50.  ])),
                        (np.array([1.   ,  0.001]),
                        np.array([1, 120]))
                        ])
@pytest.mark.parametrize("x",[
                        (np.array([100, 0.00001])),
                        (np.array([0, -.45])),
                        (np.array([-3.14159265, 2.23606798]))])
def test_lognormal_logpdf(mean,std, x ):

    # CUQI lognormal x1,x2
    R = np.array([[1, 0], [0, 1]])
    LND = cuqi.distribution.Lognormal(mean, std**2*R)
    
    # Scipy lognormal for x1
    x_1_pdf_scipy = sp.stats.lognorm.pdf(x[0], s = std[0], scale= np.exp(mean[0]))

    # Scipy lognormal for x2
    x_2_pdf_scipy = sp.stats.lognorm.pdf(x[1], s = std[1], scale= np.exp(mean[1]))
    
    # x1 and x2 are independent 
    assert(np.isclose(LND.pdf(x), x_1_pdf_scipy*x_2_pdf_scipy))
    assert(np.isclose(LND.logpdf(x), np.log(x_1_pdf_scipy*x_2_pdf_scipy)))

@pytest.mark.parametrize("mean, std, R",[
                        (np.array([0, 0]),
                        np.array([1, 1]),
                        np.array([[1, .3], [.3, 1]])),
                        (np.array([-3.14159265,  2.23606798]),
                        np.array([3.14159265, 50.  ]),
                        np.array([[1, 0], [0, 1]])),
                        (np.array([1.   ,  0.001]),
                        np.array([1, 120]),
                        np.array([[2, 2], [2, 14]]))
                        ])
@pytest.mark.parametrize("val",[
                        (np.array([100, 0.1])),
                        (np.array([-10, .1])),
                        (np.array([0.1, .45])),
                        (np.array([3.14159265, 2.23606798]))])
def test_gradient_lognormal_as_prior(mean, std, R, val):
    # This test verifies the lognormal distribution gradient correctness
    # (when the mean is a numpy.ndarray) by comparing it with a finite
    # difference approximation of the gradient.

    # ------------------- 1. Create lognormal distribution --------------------
    Sigma = np.diag(std)@R@np.diag(std) # std**2*R
    LND = cuqi.distribution.Lognormal(mean, Sigma)

    # -------------- 2. Create the finite difference gradient -----------------
    eps = 0.000001
    FD_gradient = np.empty(LND.dim)

    for i in range(LND.dim):
        # compute the ith component of the gradient
        eps_vec = np.zeros(LND.dim)
        eps_vec[i] = eps
        val_plus_eps = val + eps_vec
        FD_gradient[i] = (LND.logpdf(val_plus_eps) - LND.logpdf(val))/eps 

    # ---------------- 3. Verify correctness of the gradient ------------------
    assert(np.all(np.isclose(FD_gradient, LND.gradient(val))) or
           (np.all(np.isnan(FD_gradient)) and np.all(np.isnan(LND.gradient(val)))) )

@pytest.mark.parametrize("std, R",[
                        (
                        3,
                        np.array([[1, .3, 0], 
                                  [.3, 1, .3], 
                                  [0, .3, 1]])),
                        (
                        12,
                        np.array([[1, 0, 0], 
                                  [0, 1, 0], 
                                  [0, 0, 1]]))
                        ])
@pytest.mark.parametrize("val",[
                        (np.array([10, 0.1, 3])),
                        (np.array([0.1, .45, 6]))])
@pytest.mark.parametrize("x",[
                        (np.array([1, -0.1])),
                        (np.array([-3.14159265, 2.23606798]))])
                      
def test_gradient_lognormal_as_likelihood(std, R, val, x):
    # This test verifies the lognormal distribution gradient correctness
    # (when the mean is a cuqi.Model) by comparing it with a finite
    # difference approximation of the gradient.

    # ------------------------ 1. Create a cuqi.model -------------------------
    domain_geometry = 2
    range_geometry = 3

    # model's forward function
    def forward(x):
        return np.array([np.exp(x[0]) + x[1],
             np.exp(2*x[1]) + 3*x[0],
             x[0]+2*x[1]])
    
    # model's gradient
    def gradient(val, x=None):
        jac = np.zeros((range_geometry, domain_geometry))
        jac[0,0] =  np.exp(x[0])#d f1/ d x1
        jac[0,1] =  1 #d f1/ d x2
    
        jac[1,0] =  3 #d f2/ d x1
        jac[1,1] =  2*np.exp(2*x[1]) #d f2/ d x2
    
        jac[2,0] =  1 #d f2/ d x1
        jac[2,1] =  2 #d f2/ d x2
    
        return jac.T@val
    
    # create the cuqi.Model
    model = cuqi.model.Model(forward=forward, 
                             domain_geometry=domain_geometry, 
    			 range_geometry=range_geometry)
    model.gradient = gradient
    
    # ------------------- 2. Create lognormal distribution --------------------
    LND = cuqi.distribution.Lognormal(model, std**2*R)
    
    # -------------- 3. Create the finite difference gradient -----------------
    eps = 0.000001
    FD_gradient = np.empty(domain_geometry)
 
    for i in range(domain_geometry):
        eps_vec = np.zeros(domain_geometry)
        eps_vec[i] = eps
        x_plus_eps = x + eps_vec
        FD_gradient[i] = (LND(x=x_plus_eps).logpdf(val) - LND(x=x).logpdf(val))/eps
    
    # ---------------- 4. Verify correctness of the gradient ------------------
    assert(np.all(np.isclose(FD_gradient, LND.gradient(val, x=x))))

def test_beta(): #TODO. Make more tests
    alpha = 2; beta = 5; x=0.5

    # Create beta distribution
    BD = cuqi.distribution.Beta(alpha, beta)

    # Gamma function
    gamma = lambda x: sp.special.gamma(x)

    # Direct PDF formula for Beta
    def my_pdf( x, a, b):
        if np.any(x<=0) or np.any(x>=1) or np.any(a<=0) or np.any(b<=0):
            val = x*0
        else:
            val = gamma(a+b)/(gamma(a)*gamma(b)) * x**(a-1) * (1-x)**(b-1)
        return val

    # PDF
    assert np.allclose(BD.pdf(x), my_pdf(x, BD.alpha, BD.beta))

    # CDF
    assert np.allclose(BD.cdf(x),np.prod(sp.stats.beta.cdf(x, a=alpha, b=beta)))

    # logpdf
    assert np.allclose(BD.logpdf(x), np.log(my_pdf(x, alpha, beta)))
    
    # GRADIENT
    FD_gradient = cuqi.utilities.approx_gradient(BD.logpdf, x, epsilon=0.000000001)
    assert np.allclose(BD.gradient(x),FD_gradient,rtol=1e-3) or (np.all(np.isnan(FD_gradient)) and np.all(np.isnan(BD.gradient(x))))

# Fixture for beta distribution where beta is a likelihood
@pytest.fixture
def beta_likelihood():
    # simple forward model
    A = cuqi.model.Model(
        lambda x: x**2,
        range_geometry=1,
        domain_geometry=1,
        gradient=lambda direction, x: 2*x*direction)
    
    # set a gaussian prior
    x = cuqi.distribution.Gaussian(0, 1)
    # Beta data distribution
    y = cuqi.distribution.Beta(A(x),1)
    # set the observed data
    y=y(y=0.5)
    return y

def test_gradient_for_Beta_as_likelihood_raises_error(beta_likelihood):
    """Test computing the gradient of the Beta distribution as a likelihood
    raises a NotImplementedError"""

    with pytest.raises(NotImplementedError, 
                       match=r"Gradient is not implemented for CUQI Beta."):
        beta_likelihood.gradient(1)

@pytest.mark.parametrize("C",[1, np.ones(5), np.eye(5), sps.eye(5), sps.diags(np.ones(5))])
def test_Gaussian_Cov_sample(C):
    x = cuqi.distribution.Gaussian(np.zeros(5), np.pi*C)
    rng = np.random.RandomState(0)
    samples = x.sample(rng=rng)
    assert np.allclose(samples, np.array([3.12670137, 0.70926018, 1.73476791, 3.97187978, 3.31016035]))


@pytest.mark.parametrize("dist",
                         [cuqi.distribution.Gaussian(np.zeros(2), np.eye(2)),
                          cuqi.distribution.Beta(np.ones(2)*2, 5),
                          cuqi.distribution.Lognormal(np.ones(2)*.1, 4),
                          cuqi.distribution.Gaussian(np.zeros(2),
                                                     np.array([[1.0, 0.7],
                                                               [0.7,  1.]])),
                          cuqi.distribution.GMRF(np.zeros(2), 0.1)])
def test_enable_FD_gradient_distributions(dist):
    """Test that the distribution FD gradient is close to the exact gradient"""
    x = np.array([0.1, 0.3])

    # Exact gradient
    g_exact = dist.gradient(x)

    # FD gradient
    dist.enable_FD()
    g_FD = dist.gradient(x)

    # Assert that the FD gradient is close to the exact gradient
    # but not exactly equal to it
    assert np.allclose(g_exact, g_FD) and np.all(g_exact != g_FD)


@pytest.mark.parametrize("x",
                         [cuqi.distribution.Gaussian(np.zeros(6), np.eye(6)),
                          cuqi.distribution.Beta(np.ones(6)*2, 5),
                          cuqi.distribution.Lognormal(np.ones(6)*.1, 4),
                          cuqi.distribution.GMRF(np.zeros(6), 0.1)])
@pytest.mark.parametrize("y",
                         [cuqi.distribution.Gaussian(np.zeros(6), np.eye(6)),
                          cuqi.distribution.Lognormal(np.zeros(6), 4)])
@pytest.mark.parametrize("x_i", [np.array([0.1, 0.3, 6, 12, 1, 2]),
                                 np.array([0.1, 0.3, 0.5, 6, 3, 1])])
def test_enable_FD_gradient_posterior(x, y, x_i):
    """ Test that the posterior exact gradient and FD gradient are close."""

    # Create a model
    model = cuqi.testproblem.Deconvolution1D(dim=6).model

    # Create likelihood
    y.mean = model
    data = y(x_i).sample()
    likelihood = y(y=data)

    # Create posterior
    posterior = cuqi.distribution.Posterior(likelihood, x)

    # Compute exact gradient
    g_exact = posterior.gradient(x_i)

    # Compute FD gradient
    posterior.enable_FD(1e-7)
    g_FD = posterior.gradient(x_i)

    # Assert that the exact and FD gradient are close,
    # but not exactly equal (since we use a different method)
    assert np.allclose(g_exact, g_FD) and np.all(g_exact != g_FD)\
        or (np.all(np.isnan(g_exact)) and np.all(np.isnan(g_FD)))

def test_Cauchy_against_scipy():
    """ Test Cauchy distribution logpdf, cdf, pdf, gradient """

    x = cuqi.distribution.Cauchy(np.random.randn(5), np.abs(np.random.rand(5)))

    val = np.random.randn(5)

    # Test logpdf
    assert np.allclose(x.logpdf(val), np.sum(scipy_stats.cauchy.logpdf(val, loc=x.location, scale=x.scale)))

    # Test pdf
    assert np.allclose(x.pdf(val), np.prod(scipy_stats.cauchy.pdf(val, loc=x.location, scale=x.scale)))

    # Test cdf
    assert np.allclose(x.cdf(val), np.sum(scipy_stats.cauchy.cdf(val, loc=x.location, scale=x.scale)))

    # Test gradient
    assert np.allclose(x.gradient(val), cuqi.utilities.approx_derivative(x.logpdf, val))

def test_Cauchy_out_of_range_values():
    """ Test that the logpdf is -inf for values outside the support """

    x = cuqi.distribution.Cauchy(np.random.randn(5), np.abs(np.random.rand(5)))

    x.scale[0] = 0 # This is not a valid scale

    val = np.random.randn(5)

    # Test logpdf
    assert np.allclose(x.logpdf(val), -np.inf)

def test_Gaussian_sqrtprec_must_be_square():
    """ Test Gaussian distribution raises ValueError if sqrtprec (sparse and dense) is not square """
    N = 10; M = 5

    # Create non square sqrtprec as sparse (later converted to dense)
    sqrtprec = sp.sparse.csr_matrix(np.random.randn(N, M))

    with pytest.raises(ValueError, match="sqrtprec must be square"):
        cuqi.distribution.Gaussian(mean = np.zeros(N), sqrtprec = sqrtprec)

    with pytest.raises(ValueError, match="sqrtprec must be square"):
        cuqi.distribution.Gaussian(mean = np.zeros(N), sqrtprec = sqrtprec.todense())

def test_Gaussian_from_sparse_sqrtprec():
    """ Test Gaussian distribution from sparse sqrtprec is equal to dense sqrtprec """
    N = 10; M = 5

    sqrtprec = sp.sparse.spdiags(np.random.randn(N), 0, N, N)

    y_from_sparse = cuqi.distribution.Gaussian(mean = np.zeros(N), sqrtprec = sqrtprec)
    y_from_dense = cuqi.distribution.Gaussian(mean = np.zeros(N), sqrtprec = sqrtprec.todense())

    assert y_from_dense.logpdf(np.ones(N)) == y_from_sparse.logpdf(np.ones(N))

def test_Gaussian_from_linear_operator_sqrtprec():
    """ Test Gaussian distribution from LinearOperator sqrtprec is equal to dense sqrtprec """
    N = 10; M = 5

    sqrtprec = sp.sparse.spdiags(np.random.randn(N), 0, N, N)
    prec = sqrtprec.todense()@sqrtprec.todense().T

    def matvec(x):
        return sqrtprec @ x
    def rmatvec(x):
        return sqrtprec.T @ x
    
    sqrtprec_operator = sp.sparse.linalg.LinearOperator((N, N), matvec=matvec, rmatvec=rmatvec)
    sqrtprec_operator.logdet = -np.log(nplinalg.det(prec))

    y_from_sparse = cuqi.distribution.Gaussian(mean = np.zeros(N), sqrtprec = sqrtprec_operator)
    y_from_dense = cuqi.distribution.Gaussian(mean = np.zeros(N), sqrtprec = sqrtprec.todense())

    assert np.allclose(y_from_dense.logpdf(np.ones(N)), y_from_sparse.logpdf(np.ones(N)))

@pytest.mark.parametrize("alpha, beta, gamma, expected",[
                        (1.0, 2.0, 3.0, [[0.77974597], [0.77361298], [0.5422682 ], [0.81054637], [1.35205349]]),
                        (128.0, 3.0, -4.0, [[4.80290413], [4.40134124], [4.56151431], [4.2430851 ], [4.50618522]])
                        ])
def test_MHN_sample_regression(alpha, beta, gamma, expected):
    rng = np.random.RandomState(0)
    dist = cuqi.distribution.ModifiedHalfNormal(alpha, beta, gamma)
    samples = dist.sample(5,rng=rng).samples
    assert np.allclose( samples, np.array(expected))

@pytest.mark.parametrize("alpha, beta, gamma, expected_logpdf, expected_gradient",[
                        (1.0, 2.0, 3.0, -65.0, [[-1.0], [-5.0], [-9.0], [-13.0], [-17.0]]),
                        (64.0, 3.0, -4.0, 76.6119797952689, [[53.0], [15.5], [-1.0], [-12.25], [-21.4]])
                        ])
def test_MHN_regression(alpha, beta, gamma, expected_logpdf, expected_gradient):
    dist = cuqi.distribution.ModifiedHalfNormal(alpha, beta, gamma)
    logpdf = dist.logpdf(np.array([1.0, 2.0, 3.0, 4.0, 5.0]))
    gradient = dist._gradient(np.array([1.0, 2.0, 3.0, 4.0, 5.0]))
    assert np.allclose( logpdf, np.array(expected_logpdf))
    assert np.allclose( gradient, np.array(expected_gradient))

def test_Smoothed_Laplace():
    """ Test Smoothed Laplace distribution logpdf and gradient """

    location = np.array([1, 2])
    scale = np.array([1, 2])
    scalar_laplace_0 = cuqi.distribution.Laplace(location[0], scale[0])

    scalar_smoothed_laplace_0 = cuqi.distribution.SmoothedLaplace(location[0], scale[0], 1e-8)
    scalar_smoothed_laplace_1 = cuqi.distribution.SmoothedLaplace(location[1], scale[1], 1e-8)
    vector_smoothed_laplace = cuqi.distribution.SmoothedLaplace(location, scale, 1e-8)

    x = np.array([3, 4])

    # logpdf (scalar Laplace vs scalar Smoothed Laplace)
    assert np.allclose(scalar_laplace_0.logpdf([x[0]]), scalar_smoothed_laplace_0.logpdf(x[0]))

    # logpdf (scalar Smoothed Laplace * scalar Smoothed Laplace vs vector Smoothed Laplace)
    assert np.allclose(scalar_smoothed_laplace_0.logpdf(x[0])+scalar_smoothed_laplace_1.logpdf(x[1]),
                       vector_smoothed_laplace.logpdf(x))
    
    # gradient (scalar Smoothed Laplace vs analytical)
    assert np.allclose(scalar_smoothed_laplace_0.gradient(x[0]), -1/scale[0])

    # gradient (vector Smoothed Laplace vs analytical)
    assert np.allclose(vector_smoothed_laplace.gradient(x), -1/scale)



import scipy as sp
from scipy.linalg.interpolative import estimate_spectral_norm
from scipy.sparse.linalg import LinearOperator as scipyLinearOperator
import numpy as np
import cuqi
from cuqi.solver import CGLS, FISTA, ADMM, ScipyLinearLSQ, ScipyMinimizer
from cuqi.experimental.mcmc import Sampler


class LinearRTO(Sampler):
    """
    Linear RTO (Randomize-Then-Optimize) sampler.

    Samples posterior related to the inverse problem with Gaussian likelihood and prior, and where the forward model is linear or more generally affine.

    Parameters
    ------------
    target : `cuqi.distribution.Posterior`, `cuqi.distribution.MultipleLikelihoodPosterior` or 5-dimensional tuple.
        If target is of type cuqi.distribution.Posterior or cuqi.distribution.MultipleLikelihoodPosterior, it represents the posterior distribution.
        If target is a 5-dimensional tuple, it assumes the following structure:
        (data, model, L_sqrtprec, P_mean, P_sqrtrec)
        
        Here:
        data: is a m-dimensional numpy array containing the measured data.
        model: is a m by n dimensional matrix, AffineModel or LinearModel representing the forward model.
        L_sqrtprec: is the squareroot of the precision matrix of the Gaussian likelihood.
        P_mean: is the prior mean.
        P_sqrtprec: is the squareroot of the precision matrix of the Gaussian mean.

    initial_point : `np.ndarray` 
        Initial point for the sampler. *Optional*.

    maxit : int
        Maximum number of iterations of the inner CGLS solver. *Optional*.

    tol : float
        Tolerance of the inner CGLS solver. *Optional*.

    inner_initial_point : string or np.ndarray or cuqi.array.CUQIArray
        Initial point for the inner optimization problem. Can be "previous_sample" (default), "MAP", or a specific numpy or cuqi array. *Optional*.
        
    callback : callable, optional
        A function that will be called after each sampling step. It can be useful for monitoring the sampler during sampling.
        The function should take three arguments: the sampler object, the index of the current sampling step, the total number of requested samples. The last two arguments are integers. An example of the callback function signature is: `callback(sampler, sample_index, num_of_samples)`.
        
    """
    def __init__(self, target=None, initial_point=None, maxit=10, tol=1e-6, inner_initial_point="previous_sample", **kwargs):

        super().__init__(target=target, initial_point=initial_point, **kwargs)

        # Other parameters
        self.maxit = maxit
        self.tol = tol
        self.inner_initial_point = inner_initial_point

    def _initialize(self):
        self._precompute()
        self._compute_map()

    @property
    def inner_initial_point(self):
        if isinstance(self._inner_initial_point, str):
            if self._inner_initial_point == "previous_sample":
                return self.current_point
            elif self._inner_initial_point == "map":
                return self._map
        else:
            return self._inner_initial_point

    @inner_initial_point.setter
    def inner_initial_point(self, value):
        is_correct_string = (isinstance(value, str) and
                             (value.lower() == "previous_sample" or
                              value.lower() == "map"))
        if is_correct_string:
            self._inner_initial_point = value.lower()
        elif isinstance(value, (np.ndarray, cuqi.array.CUQIarray)):
            self._inner_initial_point = value
        else:
            raise ValueError("Invalid value for inner_initial_point. Choose either 'previous_sample', 'MAP', or provide a numpy array/cuqi array.")

    @property
    def prior(self):
        return self.target.prior

    @property
    def likelihood(self):
        return self.target.likelihood
    
    @property
    def likelihoods(self):
        if isinstance(self.target, cuqi.distribution.Posterior):
            return [self.target.likelihood]
        elif isinstance(self.target, cuqi.distribution.MultipleLikelihoodPosterior):
            return self.target.likelihoods

    @property
    def model(self):
        return self.target.model 

    @property
    def models(self):
        if isinstance(self.target, cuqi.distribution.Posterior):
            return [self.target.model]
        elif isinstance(self.target, cuqi.distribution.MultipleLikelihoodPosterior):
            return self.target.models    

    def _compute_map(self):
        sim = CGLS(self.M, self.b_tild, self.current_point, self.maxit, self.tol)            
        self._map, _ = sim.solve()

    def _precompute(self):
        L1 = [likelihood.distribution.sqrtprec for likelihood in self.likelihoods]
        L2 = self.prior.sqrtprec
        L2mu = self.prior.sqrtprecTimesMean

        # pre-computations
        self.n = self.prior.dim
        self.b_tild = np.hstack([L@(likelihood.data - model._shift) for (L, likelihood, model) in zip(L1, self.likelihoods, self.models)]+ [L2mu]) # With shift from AffineModel
        callability = [callable(likelihood.model) for likelihood in self.likelihoods]
        notcallability = [not c for c in callability]
        if all(notcallability):
            self.M = sp.sparse.vstack([L@likelihood.model for (L, likelihood) in zip(L1, self.likelihoods)] + [L2])
        elif all(callability):
            # in this case, model is a function doing forward and backward operations
            def M(x, flag):
                if flag == 1:
                    out1 = [L @ likelihood.model._forward_func_no_shift(x) for (L, likelihood) in zip(L1, self.likelihoods)] # Use forward function which excludes shift
                    out2 = L2 @ x
                    out  = np.hstack(out1 + [out2])
                elif flag == 2:
                    idx_start = 0
                    idx_end = 0
                    out1 = np.zeros(self.n)
                    for likelihood in self.likelihoods:
                        idx_end += len(likelihood.data)
                        out1 += likelihood.model._adjoint_func_no_shift(likelihood.distribution.sqrtprec.T@x[idx_start:idx_end])
                        idx_start = idx_end
                    out2 = L2.T @ x[idx_end:]
                    out  = out1 + out2                
                return out   
            self.M = M  
        else:
            raise TypeError("All likelihoods need to be callable or none need to be callable.")

    def step(self):
        y = self.b_tild + np.random.randn(len(self.b_tild))
        sim = CGLS(self.M, y, self.inner_initial_point, self.maxit, self.tol)            
        self.current_point, _ = sim.solve()
        acc = 1
        return acc

    def tune(self, skip_len, update_count):
        pass
    
    def validate_target(self):
        # Check target type
        if not isinstance(self.target, (cuqi.distribution.Posterior, cuqi.distribution.MultipleLikelihoodPosterior)):
            raise ValueError(f"To initialize an object of type {self.__class__}, 'target' need to be of type 'cuqi.distribution.Posterior' or 'cuqi.distribution.MultipleLikelihoodPosterior'.")       

        # Check Linear model and Gaussian likelihood(s)
        if isinstance(self.target, cuqi.distribution.Posterior):
            if not isinstance(self.model, cuqi.model.AffineModel):
                raise TypeError("Model needs to be linear or more generally affine")

            if not hasattr(self.likelihood.distribution, "sqrtprec"):
                raise TypeError("Distribution in Likelihood must contain a sqrtprec attribute")
            
        elif isinstance(self.target, cuqi.distribution.MultipleLikelihoodPosterior): # Elif used for further alternatives, e.g., stacked posterior
            for likelihood in self.likelihoods:
                if not isinstance(likelihood.model, cuqi.model.AffineModel):
                    raise TypeError("Model needs to be linear or more generally affine")

                if not hasattr(likelihood.distribution, "sqrtprec"):
                    raise TypeError("Distribution in Likelihood must contain a sqrtprec attribute")
        
        # Check Gaussian prior
        if not hasattr(self.prior, "sqrtprec"):
            raise TypeError("prior must contain a sqrtprec attribute")

        if not hasattr(self.prior, "sqrtprecTimesMean"):
            raise TypeError("Prior must contain a sqrtprecTimesMean attribute")
    
    def _get_default_initial_point(self, dim):
        """ Get the default initial point for the sampler. Defaults to an array of zeros. """
        return np.zeros(dim)

class RegularizedLinearRTO(LinearRTO):
    """
    Regularized Linear RTO (Randomize-Then-Optimize) sampler.

    Samples posterior related to the inverse problem with Gaussian likelihood and implicit Gaussian prior, and where the forward model is Linear.
    The sampler works by repeatedly solving regularized linear least squares problems for perturbed data.
    The solver for these optimization problems is chosen based on how the regularized is provided in the implicit Gaussian prior.
    Currently we use the following solvers:
    FISTA: [1] Beck, Amir, and Marc Teboulle. "A fast iterative shrinkage-thresholding algorithm for linear inverse problems." SIAM journal on imaging sciences 2.1 (2009): 183-202.
           Used when prior.proximal is callable.
    ADMM:  [2] Boyd et al. "Distributed optimization and statistical learning via the alternating direction method of multipliers."Foundations and Trends® in Machine learning, 2011.
           Used when prior.proximal is a list of penalty terms.
    ScipyLinearLSQ: Wrapper for Scipy's lsq_linear for the Trust Region Reflective algorithm. Optionally used when the constraint is either "nonnegativity" or "box".
    ScipyMinimizer: Wrapper for Scipy's minimize. Optionally used when the constraint is either "nonnegativity" or "box".

    Parameters
    ------------
    target : `cuqi.distribution.Posterior`
        See `cuqi.sampler.LinearRTO`

    initial_point : `np.ndarray` 
        Initial point for the sampler. *Optional*.

    maxit : int
        Maximum number of iterations of the FISTA/ADMM/ScipyLinearLSQ/ScipyMinimizer solver. *Optional*.

    inner_max_it : int
        Maximum number of iterations of the CGLS solver used within the ADMM solver. *Optional*.
        
    stepsize : string or float
        If stepsize is a string and equals either "automatic", then the stepsize is automatically estimated based on the spectral norm.
        If stepsize is a float, then this stepsize is used.

    penalty_parameter : int
        Penalty parameter of the ADMM solver. *Optional*.
        See [2] or `cuqi.solver.ADMM`

    abstol : float
        Absolute tolerance of the FISTA/ScipyLinearLSQ/ScipyMinimizer solver. *Optional*.
    
    inner_abstol : float
        Tolerance parameter for ScipyLinearLSQ's inner solve of the unbounded least-squares problem. *Optional*.
    
    adaptive : bool
        If True, FISTA is used as solver, otherwise ISTA is used. *Optional*.
    
    solver : string
        Options are "FISTA" (default for a single constraint or regularization), "ADMM" (default and the only option for multiple constraints or regularizations), "ScipyLinearLSQ" and "ScipyMinimizer". Note "ScipyLinearLSQ" and "ScipyMinimizer" can only be used with `RegularizedGaussian` of a single `box` or `nonnegativity` constraint. *Optional*.

    inner_initial_point : string or np.ndarray or cuqi.array.CUQIArray
        Initial point for the inner optimization problem. Can be "previous_sample" (default), "MAP", or a specific numpy or cuqi array. *Optional*.

    callback : callable, optional
        A function that will be called after each sampling step. It can be useful for monitoring the sampler during sampling.
        The function should take three arguments: the sampler object, the index of the current sampling step, the total number of requested samples. The last two arguments are integers. An example of the callback function signature is: `callback(sampler, sample_index, num_of_samples)`.
        
    """
    def __init__(self, target=None, initial_point=None, maxit=100, inner_max_it=10, stepsize="automatic", penalty_parameter=10, abstol=1e-10, adaptive=True, solver=None, inner_abstol=None, inner_initial_point="previous_sample", **kwargs):
        
        super().__init__(target=target, initial_point=initial_point, **kwargs)

        # Other parameters
        self.stepsize = stepsize
        self.abstol = abstol
        self.inner_abstol = inner_abstol
        self.adaptive = adaptive
        self.maxit = maxit
        self.inner_max_it = inner_max_it
        self.penalty_parameter = penalty_parameter
        self.solver = solver
        self.inner_initial_point = inner_initial_point

    def _initialize(self):
        super()._initialize()
        if self.solver is None:
            self.solver = "FISTA" if callable(self.proximal) else "ADMM"
        if self.solver == "FISTA":
            self._stepsize = self._choose_stepsize()
        self._compute_map_regularized()

    @property
    def solver(self):
        return self._solver

    @solver.setter
    def solver(self, value):
        if value == "ScipyLinearLSQ" or value == "ScipyMinimizer":
            if (self.target.prior.preset["constraint"] == "nonnegativity" or self.target.prior.preset["constraint"] == "box"):
                self._solver = value
            else:
                raise ValueError("ScipyLinearLSQ and ScipyMinimizer only support RegularizedGaussian with box or nonnegativity constraint.")
        else:
            self._solver = value

    @property
    def proximal(self):
        return self.target.prior.proximal
    
    def validate_target(self):
        super().validate_target()
        if not isinstance(self.target.prior, (cuqi.implicitprior.RegularizedGaussian, cuqi.implicitprior.RegularizedGMRF)):
            raise TypeError("Prior needs to be RegularizedGaussian or RegularizedGMRF")

    def _choose_stepsize(self):
        if isinstance(self.stepsize, str):
            if self.stepsize in ["automatic"]:
                if not callable(self.M):
                    M_op = scipyLinearOperator(self.M.shape, matvec = lambda v: self.M@v, rmatvec = lambda w: self.M.T@w)
                else:
                    M_op = scipyLinearOperator((len(self.b_tild), self.n), matvec = lambda v: self.M(v,1), rmatvec = lambda w: self.M(w,2))
                    
                _stepsize = 0.99/(estimate_spectral_norm(M_op)**2)
                # print(f"Estimated stepsize for regularized Linear RTO: {_stepsize}")
            else:
                raise ValueError("Stepsize choice not supported")
        else:
            _stepsize = self.stepsize
        return _stepsize

    @property
    def prior(self):
        return self.target.prior.gaussian

    def _compute_map_regularized(self):
        self._map = self._customized_step(self.b_tild, self.initial_point)

    def _customized_step(self, y, x0):
        if self.solver == "FISTA":
            sim = FISTA(self.M, y, self.proximal,
                        x0, maxit = self.maxit, stepsize = self._stepsize, abstol = self.abstol, adaptive = self.adaptive)         
        elif self.solver == "ADMM":
            sim = ADMM(self.M, y, self.proximal,
                        x0, self.penalty_parameter, maxit = self.maxit, inner_max_it = self.inner_max_it, adaptive = self.adaptive)
        elif self.solver == "ScipyLinearLSQ":
            A_op = sp.sparse.linalg.LinearOperator((sum([llh.distribution.dim for llh in self.likelihoods])+self.target.prior.dim, self.target.prior.dim),
                                    matvec=lambda x: self.M(x, 1),
                                    rmatvec=lambda x: self.M(x, 2)
                                    )
            sim = ScipyLinearLSQ(A_op, y, self.target.prior._box_bounds, 
                                    max_iter = self.maxit,
                                    lsmr_maxiter = self.inner_max_it, 
                                    tol = self.abstol,
                                    lsmr_tol = self.inner_abstol)
        elif self.solver == "ScipyMinimizer":
            # Adapt bounds format, as scipy.minimize requires a bounds format 
            # different than that in scipy.lsq_linear.
            bounds = [(self.target.prior._box_bounds[0][i], self.target.prior._box_bounds[1][i]) for i in range(self.target.prior.dim)]
            # Note that the objective function is defined as 0.5*||Mx-y||^2, 
            # and the corresponding gradient (gradfunc) is given by M^T(Mx-y).
            sim = ScipyMinimizer(lambda x: 0.5*np.sum((self.M(x, 1)-y)**2), x0, gradfunc=lambda x: self.M(self.M(x, 1) - y, 2), bounds=bounds, tol=self.abstol, options={"maxiter": self.maxit})
        else:
            raise ValueError("Choice of solver not supported.")
        
        sol, _ = sim.solve()
        return sol

    def step(self):
        y = self.b_tild + np.random.randn(len(self.b_tild))

        self.current_point = self._customized_step(y, self.inner_initial_point)

        acc = 1
        return acc


"""
Re-implementation of sampler module in a more object-oriented way.

Main changes for users
----------------------

1. Sampling API
   ^^^^^^^^^^^^

   Previously one would call the `.sample` or `sample_adapt` methods of a sampler instance at :py:mod:`cuqi.sampler` to sample from a target distribution and store the samples as the output as follows:

   .. code-block:: python

      from cuqi.sampler import MH
      from cuqi.distribution import DistributionGallery

      # Target distribution
      target = DistributionGallery("donut")

      # Set up sampler
      sampler = MH(target)

      # Sample from the target distribution (Alternatively calling sample with explicit scale parameter set in sampler)
      samples = sampler.sample_adapt(Ns=100, Nb=100) # Burn-in (Nb) removed by default

   This has now changed to to a more object-oriented API which provides more flexibility and control over the sampling process.

   For example one can now more explicitly control when the sampler is tuned (warmup) and when it is sampling with fixed parameters.

   .. code-block:: python

      from cuqi.experimental.mcmc import MH
      from cuqi.distribution import DistributionGallery

      # Target distribution
      target = DistributionGallery("donut")

      # Set up sampler
      sampler = MH(target)

      # Sample from the target distribution
      sampler.warmup(Nb=100)  # Explicit warmup (tuning) of sampler
      sampler.sample(Ns=100)  # Sampling with fixed parameters
      samples = sampler.get_samples().burnthin(Nb=100) # Getting samples and removing burn-in from warmup

   Importantly, the removal of burn-in from e.g. warmup is now a separate step that is done after the sampling process is complete.

2. Sampling API for BayesianProblem
   ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

   :py:class:`cuqi.problem.BayesianProblem` continues to have the same API for `sample_posterior` and the `UQ` method.

   There is now a flag `experimental` that can be set to `True` to use the new MCMC samplers.
   
   By default, the flag is set to `False` and the old samplers are used.

   For this more high-level interface, burn-in is automatically removed from the samples as was the case before.


3. More options for Gibbs sampling
   ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

   There are now more options for Gibbs sampling. Previously it was only possible to sample with Gibbs for samplers  :py:class:`cuqi.sampler.LinearRTO`, :py:class:`cuqi.sampler.RegularizedLinearRTO`, :py:class:`cuqi.sampler.Conjugate`, and :py:class:`cuqi.sampler.ConjugateApprox`.

   Now, it is possible to define a Gibbs sampling scheme using any sampler from the :py:mod:`cuqi.experimental.mcmc` module.

   **Example using a NUTS-within-Gibbs scheme for a 1D deconvolution problem:**

   .. code-block:: python

      import cuqi
      import numpy as np
      from cuqi.distribution import Gamma, Gaussian, GMRF, JointDistribution
      from cuqi.experimental.mcmc import NUTS, HybridGibbs, Conjugate
      from cuqi.testproblem import Deconvolution1D

      # Forward problem
      A, y_data, info = Deconvolution1D(dim=128, phantom='sinc', noise_std=0.001).get_components()

      # Bayesian Inverse Problem
      s = Gamma(1, 1e-4)
      x = GMRF(np.zeros(A.domain_dim), 50)
      y = Gaussian(A @ x, lambda s: 1 / s)

      # Posterior
      target = JointDistribution(y, x, s)(y=y_data)

      # Gibbs sampling strategy. Note we can define initial_points and various parameters for each sampler
      sampling_strategy = {
          "x": NUTS(max_depth=10, initial_point=np.zeros(A.domain_dim)),
          "s": Conjugate()
      }

      # Here we do 10 internal steps with NUTS for each Gibbs step
      num_sampling_steps = {
          "x": 10,
          "s": 1
      }

      sampler = HybridGibbs(target, sampling_strategy, num_sampling_steps)

      sampler.warmup(50)
      sampler.sample(200)
      samples = sampler.get_samples().burnthin(Nb=50)

      samples["x"].plot_ci(exact=info.exactSolution)
"""



from ._sampler import Sampler, ProposalBasedSampler
from ._langevin_algorithm import ULA, MALA, MYULA, PnPULA
from ._mh import MH
from ._pcn import PCN
from ._rto import LinearRTO, RegularizedLinearRTO
from ._cwmh import CWMH
from ._laplace_approximation import UGLA
from ._hmc import NUTS
from ._gibbs import HybridGibbs
from ._conjugate import Conjugate
from ._conjugate_approx import ConjugateApprox
from ._direct import Direct



import numpy as np
from abc import ABC, abstractmethod
import math
from cuqi.experimental.mcmc import Sampler
from cuqi.distribution import Posterior, Gaussian, Gamma, GMRF, ModifiedHalfNormal
from cuqi.implicitprior import RegularizedGaussian, RegularizedGMRF, RegularizedUnboundedUniform
from cuqi.utilities import get_non_default_args, count_nonzero, count_within_bounds, count_constant_components_1D, count_constant_components_2D, piecewise_linear_1D_DoF
from cuqi.geometry import Continuous1D, Continuous2D, Image2D

class Conjugate(Sampler):
    """ Conjugate sampler

    Sampler for sampling a posterior distribution which is a so-called "conjugate" distribution, i.e., where the likelihood and prior are conjugate to each other - denoted as a conjugate pair.   

    Currently supported conjugate pairs are:
    - (Gaussian, Gamma) where Gamma is defined on the precision parameter of the Gaussian
    - (GMRF, Gamma) where Gamma is defined on the precision parameter of the GMRF
    - (RegularizedGaussian, Gamma) with preset constraints only and Gamma is defined on the precision parameter of the RegularizedGaussian
    - (RegularizedGMRF, Gamma) with preset constraints only and Gamma is defined on the precision parameter of the RegularizedGMRF
    - (RegularizedGaussian, ModifiedHalfNormal) with most of the preset constraints and regularization
    - (RegularizedGMRF, ModifiedHalfNormal) with most of the preset constraints and regularization

    Currently the Gamma and ModifiedHalfNormal distribution must be univariate.

    A conjugate pair defines implicitly a so-called conjugate distribution which can be sampled from directly.

    The conjugate parameter is the parameter that both the likelihood and prior PDF depend on.

    For more information on conjugacy and conjugate distributions see https://en.wikipedia.org/wiki/Conjugate_prior.

    For implicit regularized Gaussians and the corresponding conjugacy relations, see:
    
    Section 3.3 from [1] Everink, Jasper M., Yiqiu Dong, and Martin S. Andersen. "Bayesian inference with projected densities." SIAM/ASA Journal on Uncertainty Quantification 11.3 (2023): 1025-1043.
    Section 4 from [2] Everink, Jasper M., Yiqiu Dong, and Martin S. Andersen. "Sparse Bayesian inference with regularized Gaussian distributions." Inverse Problems 39.11 (2023): 115004.

    """

    def _initialize(self):
        pass

    @Sampler.target.setter # Overwrite the target setter to set the conjugate pair
    def target(self, value):
        """ Set the target density. Runs validation of the target. """
        self._target = value
        if self._target is not None:
            self._set_conjugatepair()
            self.validate_target()

    def validate_target(self):
        self._ensure_target_is_posterior()
        self._conjugatepair.validate_target()
        
    def step(self):
        self.current_point = self._conjugatepair.sample()
        return 1 # Returns acceptance rate of 1

    def tune(self, skip_len, update_count):
        pass # No tuning required for conjugate sampler

    def _ensure_target_is_posterior(self):
        """ Ensure that the target is a Posterior distribution. """
        if not isinstance(self.target, Posterior):
            raise TypeError("Conjugate sampler requires a target of type Posterior")

    def _set_conjugatepair(self):
        """ Set the conjugate pair based on the likelihood and prior. This requires target to be set. """
        self._ensure_target_is_posterior()
        if isinstance(self.target.likelihood.distribution, (Gaussian, GMRF)) and isinstance(self.target.prior, Gamma):
            self._conjugatepair = _GaussianGammaPair(self.target)
        elif isinstance(self.target.likelihood.distribution, RegularizedUnboundedUniform) and isinstance(self.target.prior, Gamma):
            # Check RegularizedUnboundedUniform before RegularizedGaussian and RegularizedGMRF due to the first inheriting from the second.
            self._conjugatepair = _RegularizedUnboundedUniformGammaPair(self.target)
        elif isinstance(self.target.likelihood.distribution, (RegularizedGaussian, RegularizedGMRF)) and isinstance(self.target.prior, Gamma):
            self._conjugatepair = _RegularizedGaussianGammaPair(self.target)
        elif isinstance(self.target.likelihood.distribution, (RegularizedGaussian, RegularizedGMRF)) and isinstance(self.target.prior, ModifiedHalfNormal):
            self._conjugatepair = _RegularizedGaussianModifiedHalfNormalPair(self.target)
        else:
            raise ValueError(f"Conjugacy is not defined for likelihood {type(self.target.likelihood.distribution)} and prior {type(self.target.prior)}, in CUQIpy")

    def conjugate_distribution(self):
        return self._conjugatepair.conjugate_distribution()

    def __repr__(self):
        msg = super().__repr__()
        if hasattr(self, "_conjugatepair"):
            msg += f"\n Conjugate pair:\n\t {type(self._conjugatepair).__name__.removeprefix('_')}"
        return msg
        
class _ConjugatePair(ABC):
    """ Abstract base class for conjugate pairs (likelihood, prior) used in the Conjugate sampler. """

    def __init__(self, target):
        self.target = target

    @abstractmethod
    def validate_target(self):
        """ Validate the target distribution for the conjugate pair. """
        pass

    @abstractmethod
    def conjugate_distribution(self):
        """ Returns the posterior distribution in the form of a CUQIpy distribution """
        pass

    def sample(self):
        """ Sample from the conjugate distribution. """
        return self.conjugate_distribution().sample()


class _GaussianGammaPair(_ConjugatePair):
    """ Implementation for the Gaussian-Gamma conjugate pair."""

    def validate_target(self):
        if self.target.prior.dim != 1:
            raise ValueError("Gaussian-Gamma conjugacy only works with univariate Gamma prior")

        key_value_pairs = _get_conjugate_parameter(self.target)
        if len(key_value_pairs) != 1:
            raise ValueError(f"Multiple references to conjugate parameter {self.target.prior.name} found in likelihood. Only one occurance is supported.")
        for key, value in key_value_pairs:
            if key == "cov":
                if not _check_conjugate_parameter_is_scalar_linear_reciprocal(value):
                    raise ValueError("Gaussian-Gamma conjugate pair defined via covariance requires cov: lambda x : s/x for the conjugate parameter")
            elif key == "prec":
                if not _check_conjugate_parameter_is_scalar_linear(value):
                    raise ValueError("Gaussian-Gamma conjugate pair defined via precision requires prec: lambda x : s*x for the conjugate parameter")
            else:
                raise ValueError(f"RegularizedGaussian-ModifiedHalfNormal conjugacy does not support the conjugate parameter {self.target.prior.name} in the {key} attribute. Only cov and prec")

    def conjugate_distribution(self):
        # Extract variables
        b = self.target.likelihood.data                                 # mu
        m = len(b)                                                      # n
        Ax = self.target.likelihood.distribution.mean                   # x_i
        L = self.target.likelihood.distribution(np.array([1])).sqrtprec # L
        alpha = self.target.prior.shape                                 # alpha
        beta = self.target.prior.rate                                   # beta

        # Create Gamma distribution and sample
        return Gamma(shape=m/2 + alpha, rate=.5 * np.linalg.norm(L @ (Ax - b))**2 + beta)


class _RegularizedGaussianGammaPair(_ConjugatePair):
    """Implementation for the Regularized Gaussian-Gamma conjugate pair using the conjugacy rules from [1], Section 3.3."""

    def validate_target(self):
        if self.target.prior.dim != 1:
            raise ValueError("RegularizedGaussian-Gamma conjugacy only works with univariate ModifiedHalfNormal prior")

        # Raises error if preset is not supported
        _compute_sparsity_level(self.target)

        key_value_pairs = _get_conjugate_parameter(self.target)
        if len(key_value_pairs) != 1:
            raise ValueError(f"Multiple references to conjugate parameter {self.target.prior.name} found in likelihood. Only one occurance is supported.")
        for key, value in key_value_pairs:
            if key == "cov":
                if not _check_conjugate_parameter_is_scalar_linear_reciprocal(value):
                    raise ValueError("Regularized Gaussian-Gamma conjugacy defined via covariance requires cov: lambda x : s/x for the conjugate parameter")
            elif key == "prec":
                if not _check_conjugate_parameter_is_scalar_linear(value):
                    raise ValueError("Regularized Gaussian-Gamma conjugacy defined via precision requires prec: lambda x : s*x for the conjugate parameter")
            else:
                raise ValueError(f"RegularizedGaussian-ModifiedHalfNormal conjugacy does not support the conjugate parameter {self.target.prior.name} in the {key} attribute. Only cov and prec")

    def conjugate_distribution(self):
        # Extract variables
        b = self.target.likelihood.data                                 # mu
        m = _compute_sparsity_level(self.target)
        Ax = self.target.likelihood.distribution.mean                   # x_i
        L = self.target.likelihood.distribution(np.array([1])).sqrtprec # L
        alpha = self.target.prior.shape                                 # alpha
        beta = self.target.prior.rate                                   # beta

        # Create Gamma distribution and sample
        return Gamma(shape=m/2 + alpha, rate=.5 * np.linalg.norm(L @ (Ax - b))**2 + beta)
    
    
class _RegularizedUnboundedUniformGammaPair(_ConjugatePair):
    """Implementation for the RegularizedUnboundedUniform-ModifiedHalfNormal conjugate pair using the conjugacy rules from [2], Section 4."""

    def validate_target(self):
        if self.target.prior.dim != 1:
            raise ValueError("RegularizedUnboundedUniform-Gamma conjugacy only works with univariate Gamma prior")
        
        # Raises error if preset is not supported
        _compute_sparsity_level(self.target)

        key_value_pairs = _get_conjugate_parameter(self.target)
        if len(key_value_pairs) != 1:
            raise ValueError(f"Multiple references to conjugate parameter {self.target.prior.name} found in likelihood. Only one occurance is supported.")
        for key, value in key_value_pairs:
            if key == "strength":
                if not _check_conjugate_parameter_is_scalar_linear(value):
                    raise ValueError("RegularizedUnboundedUniform-Gamma conjugacy defined via strength requires strength: lambda x : s*x for the conjugate parameter")
            else:
                raise ValueError(f"RegularizedUnboundedUniform-Gamma conjugacy does not support the conjugate parameter {self.target.prior.name} in the {key} attribute. Only strength is supported")

    def conjugate_distribution(self):
        # Extract prior variables
        alpha = self.target.prior.shape
        beta = self.target.prior.rate

        # Compute likelihood quantities
        x = self.target.likelihood.data
        m = _compute_sparsity_level(self.target)

        reg_op = self.target.likelihood.distribution._regularization_oper
        reg_strength = self.target.likelihood.distribution(np.array([1])).strength
        fx = reg_strength*np.linalg.norm(reg_op@x, ord = 1)

        # Create Gamma distribution
        return Gamma(shape=m/2 + alpha, rate=fx + beta)
    
class _RegularizedGaussianModifiedHalfNormalPair(_ConjugatePair):
    """Implementation for the Regularized Gaussian-ModifiedHalfNormal conjugate pair using the conjugacy rules from [2], Section 4."""

    def validate_target(self):
        if self.target.prior.dim != 1:
            raise ValueError("RegularizedGaussian-ModifiedHalfNormal conjugacy only works with univariate ModifiedHalfNormal prior")
        
        # Raises error if preset is not supported
        _compute_sparsity_level(self.target)

        key_value_pairs = _get_conjugate_parameter(self.target)
        if len(key_value_pairs) != 2:
            raise ValueError(f"Incorrect number of references to conjugate parameter {self.target.prior.name} found in likelihood. Found {len(key_value_pairs)} times, but needs to occur in prec or cov, and in strength")
        for key, value in key_value_pairs:
            if key == "strength":
                if not _check_conjugate_parameter_is_scalar_linear(value):
                    raise ValueError("RegularizedGaussian-ModifiedHalfNormal conjugacy defined via strength requires strength: lambda x : s*x for the conjugate parameter")
            elif key == "prec":
                if not _check_conjugate_parameter_is_scalar_quadratic(value):
                    raise ValueError("RegularizedGaussian-ModifiedHalfNormal conjugacy defined via precision requires prec: lambda x : s*x for the conjugate parameter")
            elif key == "cov":
                if not _check_conjugate_parameter_is_scalar_quadratic_reciprocal(value):
                    raise ValueError("RegularizedGaussian-ModifiedHalfNormal conjugacy  defined via covariance requires cov: lambda x : s/x for the conjugate parameter")
            else:
                raise ValueError(f"RegularizedGaussian-ModifiedHalfNormal conjugacy does not support the conjugate parameter {self.target.prior.name} in the {key} attribute. Only cov, prec and strength are supported")


    def conjugate_distribution(self):
        # Extract prior variables
        alpha = self.target.prior.alpha
        beta = self.target.prior.beta
        gamma = self.target.prior.gamma

        # Compute likelihood variables
        x = self.target.likelihood.data
        mu = self.target.likelihood.distribution.mean
        L = self.target.likelihood.distribution(np.array([1])).sqrtprec
        
        m = _compute_sparsity_level(self.target)

        reg_op = self.target.likelihood.distribution._regularization_oper
        reg_strength = self.target.likelihood.distribution(np.array([1])).strength
        fx = reg_strength*np.linalg.norm(reg_op@x, ord = 1)

        # Compute parameters of conjugate distribution
        conj_alpha = m + alpha
        conj_beta = 0.5*np.linalg.norm(L @ (mu - x))**2 + beta
        conj_gamma = -fx + gamma

        # Create conjugate distribution
        return ModifiedHalfNormal(conj_alpha, conj_beta, conj_gamma)
    

def _compute_sparsity_level(target):
    """Computes the sparsity level in accordance with Section 4 from [2],
    this can be interpreted as the number of degrees of freedom, that is,
    the number of components n minus the dimension the of the subdifferential of the regularized.
    """
    x = target.likelihood.data

    constraint = target.likelihood.distribution.preset["constraint"]
    regularization = target.likelihood.distribution.preset["regularization"]

    # There is no reference for some of these conjugacy rules
    if constraint == "nonnegativity":
        if regularization in [None, "l1"]:
            # Number of non-zero components in x
            return count_nonzero(x)
        elif regularization == "tv" and isinstance(target.likelihood.distribution.geometry, Continuous1D):
            # Number of non-zero constant components in x
            return count_constant_components_1D(x, lower = 0.0)
        elif regularization == "tv" and isinstance(target.likelihood.distribution.geometry, (Continuous2D, Image2D)):
            # Number of non-zero constant components in x
            return count_constant_components_2D(target.likelihood.distribution.geometry.par2fun(x), lower = 0.0)
    elif constraint == "box":
        bounds = target.likelihood.distribution._box_bounds
        if regularization is None:
            # Number of components in x that are strictly between the lower and upper bound
            return count_within_bounds(x, bounds[0], bounds[1])
        elif regularization == "l1":
            # Number of components in x that are strictly between the lower and upper bound and are not zero
            return count_within_bounds(x, bounds[0], bounds[1], exception = 0.0)
        elif regularization == "tv" and isinstance(target.likelihood.distribution.geometry, Continuous1D):
            # Number of constant components in x between are strictly between the lower and upper bound
            return count_constant_components_1D(x, lower = bounds[0], upper = bounds[1])
        elif regularization == "tv" and isinstance(target.likelihood.distribution.geometry, (Continuous2D, Image2D)):
            # Number of constant components in x between are strictly between the lower and upper bound
            return count_constant_components_2D(target.likelihood.distribution.geometry.par2fun(x), lower = bounds[0], upper = bounds[1])
    elif constraint in ["increasing", "decreasing"]:
        if regularization is None:
            # Number of constant components in x
            return count_constant_components_1D(x)
        elif regularization == "l1":
            # Number of constant components in x that are not zero
            return count_constant_components_1D(x, exception = 0.0)
        elif regularization == "tv" and isinstance(target.likelihood.distribution.geometry, Continuous1D):
            # Number of constant components in x
            return count_constant_components_1D(x)
        # Increasing and decreasing cannot be done in 2D
    elif constraint in ["convex", "concave"]:
        if regularization is None:
            # Number of piecewise linear components in x
            return piecewise_linear_1D_DoF(x)
        elif regularization == "l1":
            # Number of piecewise linear components in x that are not zero
            return piecewise_linear_1D_DoF(x, exception_zero = True)
        elif regularization == "tv" and isinstance(target.likelihood.distribution.geometry, Continuous1D):
            # Number of piecewise linear components in x that are not flat
            return piecewise_linear_1D_DoF(x, exception_flat = True)
        # convex and concave has only been implemented in 1D
    elif constraint == None: 
        if regularization == "l1":
            # Number of non-zero components in x
            return count_nonzero(x)
        elif regularization == "tv" and isinstance(target.likelihood.distribution.geometry, Continuous1D):
            # Number of non-zero constant components in x
            return count_constant_components_1D(x)
        elif regularization == "tv" and isinstance(target.likelihood.distribution.geometry, (Continuous2D, Image2D)):
            # Number of non-zero constant components in x
            return count_constant_components_2D(target.likelihood.distribution.geometry.par2fun(x))

    raise ValueError("RegularizedGaussian preset constraint and regularization choice is currently not supported with conjugacy.")


def _get_conjugate_parameter(target):
    """Extract the conjugate parameter name (e.g. d), and returns the mutable variable that is defined by the conjugate parameter, e.g. cov and its value e.g. lambda d:1/d"""
    par_name = target.prior.name
    mutable_likelihood_vars  = target.likelihood.distribution.get_mutable_variables()

    found_parameter_pairs = []

    for var_key in mutable_likelihood_vars:
        attr = getattr(target.likelihood.distribution, var_key)
        if callable(attr) and par_name in get_non_default_args(attr):
            found_parameter_pairs.append((var_key, attr))
    if len(found_parameter_pairs) == 0:
        raise ValueError(f"Unable to find conjugate parameter {par_name} in likelihood function for conjugate sampler with target {target}")
    return found_parameter_pairs

def _check_conjugate_parameter_is_scalar_identity(f):
    """Tests whether a function (scalar to scalar) is the identity (lambda x: x)."""
    test_values = [1.0, 10.0, 100.0]
    return all(np.allclose(f(x), x) for x in test_values)

def _check_conjugate_parameter_is_scalar_reciprocal(f):
    """Tests whether a function (scalar to scalar) is the reciprocal (lambda x : 1.0/x)."""
    return all(math.isclose(f(x), 1.0 / x) for x in [1.0, 10.0, 100.0])

def _check_conjugate_parameter_is_scalar_linear(f):
    """
        Tests whether a function (scalar to scalar) is linear (lambda x: s*x for some s).
        The tests checks whether the function is zero and some finite differences are constant.
    """
    test_values = [1.0, 10.0, 100.0]
    h = 1e-2
    finite_diffs = [(f(x + h*x)-f(x))/(h*x) for x in test_values]
    return np.isclose(f(0.0), 0.0) and all(np.allclose(c, finite_diffs[0]) for c in finite_diffs[1:])

def _check_conjugate_parameter_is_scalar_linear_reciprocal(f):
    """
        Tests whether a function (scalar to scalar) is a constant times the inverse of the input (lambda x: s/x for some s).
        The tests checks whether the the reciprocal of the function has constant finite differences.
    """
    g = lambda x : 1.0/f(x)
    test_values = [1.0, 10.0, 100.0]
    h = 1e-2
    finite_diffs = [(g(x + h*x)-g(x))/(h*x) for x in test_values]
    return all(np.allclose(c, finite_diffs[0]) for c in finite_diffs[1:])

def _check_conjugate_parameter_is_scalar_quadratic(f):
    """
        Tests whether a function (scalar to scalar) is linear (lambda x: s*x**2 for some s).
        The tests checks whether the function divided by the parameter is linear
    """
    return _check_conjugate_parameter_is_scalar_linear(lambda x: f(x)/x if x != 0.0 else f(0.0))

def _check_conjugate_parameter_is_scalar_quadratic_reciprocal(f):
    """
        Tests whether a function (scalar to scalar) is linear (lambda x: s*x**-2 for some s).
        The tests checks whether the function divided by the parameter is the reciprical of a linear function.
    """
    return _check_conjugate_parameter_is_scalar_linear_reciprocal(lambda x: f(x)/x)
